# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak

# https://github.com/scikit-hep/awkward-1.0/issues/459#issuecomment-694941328
#
# So the rules would be,
#    * if arrays have different `__array__` or `__record__` parameters, they are not equal;
#    * if they otherwise have different parameters, the types can be equal, but merging
#      (concatenation, option-simplify, or union-simplify) removes parameters other than
#      `__array__` and `__record__`.


def test_0459_types():
    plain_plain = ak.highlevel.Array([[0.0, 1.1, 2.2, 3.3, 4.4]])
    array_plain = ak.operations.with_parameter(plain_plain, "__list__", "zoinks")
    plain_isdoc = ak.operations.with_parameter(
        plain_plain, "__doc__", "This is a zoink."
    )
    array_isdoc = ak.operations.with_parameter(
        array_plain, "__doc__", "This is a zoink."
    )

    assert ak.operations.parameters(plain_plain) == {}
    assert ak.operations.parameters(array_plain) == {"__list__": "zoinks"}
    assert ak.operations.parameters(plain_isdoc) == {"__doc__": "This is a zoink."}
    assert ak.operations.parameters(array_isdoc) == {
        "__list__": "zoinks",
        "__doc__": "This is a zoink.",
    }

    assert ak.operations.type(plain_plain) == ak.operations.type(plain_plain)
    assert ak.operations.type(array_plain) == ak.operations.type(array_plain)
    assert ak.operations.type(plain_isdoc) == ak.operations.type(plain_isdoc)
    assert ak.operations.type(array_isdoc) == ak.operations.type(array_isdoc)

    assert ak.operations.type(plain_plain) != ak.operations.type(array_plain)
    assert ak.operations.type(array_plain) != ak.operations.type(plain_plain)

    assert ak.operations.type(plain_plain) == ak.operations.type(plain_isdoc)
    assert ak.operations.type(plain_isdoc) == ak.operations.type(plain_plain)

    assert ak.operations.type(array_plain) == ak.operations.type(array_isdoc)
    assert ak.operations.type(array_isdoc) == ak.operations.type(array_plain)

    assert ak.operations.type(plain_isdoc) != ak.operations.type(array_isdoc)
    assert ak.operations.type(array_isdoc) != ak.operations.type(plain_isdoc)

    assert array_plain.layout.parameters == {"__list__": "zoinks"}
    assert ak.operations.without_parameters(array_plain).layout.parameters == {}
    assert plain_isdoc.layout.parameters == {"__doc__": "This is a zoink."}
    assert ak.operations.without_parameters(plain_isdoc).layout.parameters == {}
    assert array_isdoc.layout.parameters == {
        "__list__": "zoinks",
        "__doc__": "This is a zoink.",
    }
    assert ak.operations.without_parameters(array_isdoc).layout.parameters == {}


def test_0459():
    plain_plain = ak.highlevel.Array([[0.0, 1.1, 2.2, 3.3, 4.4]])
    array_plain = ak.operations.with_parameter(plain_plain, "__list__", "zoinks")
    plain_isdoc = ak.operations.with_parameter(
        plain_plain, "__doc__", "This is a zoink."
    )
    array_isdoc = ak.operations.with_parameter(
        array_plain, "__doc__", "This is a zoink."
    )

    assert ak.operations.parameters(plain_plain) == {}
    assert ak.operations.parameters(array_plain) == {"__list__": "zoinks"}
    assert ak.operations.parameters(plain_isdoc) == {"__doc__": "This is a zoink."}
    assert ak.operations.parameters(array_isdoc) == {
        "__list__": "zoinks",
        "__doc__": "This is a zoink.",
    }

    assert (
        ak.operations.parameters(ak.operations.concatenate([plain_plain, plain_plain]))
        == {}
    )
    assert ak.operations.parameters(
        ak.operations.concatenate([array_plain, array_plain])
    ) == {"__list__": "zoinks"}
    assert ak.operations.parameters(
        ak.operations.concatenate([plain_isdoc, plain_isdoc])
    ) == {"__doc__": "This is a zoink."}
    assert ak.operations.parameters(
        ak.operations.concatenate([array_isdoc, array_isdoc])
    ) == {
        "__list__": "zoinks",
        "__doc__": "This is a zoink.",
    }

    assert isinstance(
        ak.operations.concatenate([plain_plain, plain_plain]).layout,
        ak.contents.ListOffsetArray,
    )
    assert isinstance(
        ak.operations.concatenate([array_plain, array_plain]).layout,
        ak.contents.ListOffsetArray,
    )
    assert isinstance(
        ak.operations.concatenate([plain_isdoc, plain_isdoc]).layout,
        ak.contents.ListOffsetArray,
    )
    assert isinstance(
        ak.operations.concatenate([array_isdoc, array_isdoc]).layout,
        ak.contents.ListOffsetArray,
    )

    assert (
        ak.operations.parameters(ak.operations.concatenate([plain_plain, array_plain]))
        == {}
    )
    assert (
        ak.operations.parameters(ak.operations.concatenate([plain_isdoc, array_isdoc]))
        == {}
    )
    assert (
        ak.operations.parameters(ak.operations.concatenate([array_plain, plain_plain]))
        == {}
    )
    assert (
        ak.operations.parameters(ak.operations.concatenate([array_isdoc, plain_isdoc]))
        == {}
    )

    assert isinstance(
        ak.operations.concatenate([plain_plain, array_plain]).layout,
        ak.contents.UnionArray,
    )
    assert isinstance(
        ak.operations.concatenate([plain_isdoc, array_isdoc]).layout,
        ak.contents.UnionArray,
    )
    assert isinstance(
        ak.operations.concatenate([array_plain, plain_plain]).layout,
        ak.contents.UnionArray,
    )
    assert isinstance(
        ak.operations.concatenate([array_isdoc, plain_isdoc]).layout,
        ak.contents.UnionArray,
    )

    assert (
        ak.operations.parameters(ak.operations.concatenate([plain_plain, plain_isdoc]))
        == {}
    )
    assert ak.operations.parameters(
        ak.operations.concatenate([array_plain, array_isdoc])
    ) == {"__list__": "zoinks"}
    assert (
        ak.operations.parameters(ak.operations.concatenate([plain_isdoc, plain_plain]))
        == {}
    )
    assert ak.operations.parameters(
        ak.operations.concatenate([array_isdoc, array_plain])
    ) == {"__list__": "zoinks"}

    assert isinstance(
        ak.operations.concatenate([plain_plain, plain_isdoc]).layout,
        ak.contents.ListOffsetArray,
    )
    assert isinstance(
        ak.operations.concatenate([array_plain, array_isdoc]).layout,
        ak.contents.ListOffsetArray,
    )
    assert isinstance(
        ak.operations.concatenate([plain_isdoc, plain_plain]).layout,
        ak.contents.ListOffsetArray,
    )
    assert isinstance(
        ak.operations.concatenate([array_isdoc, array_plain]).layout,
        ak.contents.ListOffsetArray,
    )


def test_0522():
    content1 = ak.highlevel.Array([0.0, 1.1, 2.2, 3.3, 4.4]).layout
    content2 = ak.highlevel.Array([[0], [100], [200], [300], [400]]).layout
    tags = ak.index.Index8(np.array([0, 0, 0, 1, 1, 0, 0, 1, 1, 1], np.int8))
    index = ak.index.Index64(np.array([0, 1, 2, 0, 1, 3, 4, 2, 3, 4], np.int64))
    unionarray = ak.highlevel.Array(
        ak.contents.UnionArray(tags, index, [content1, content2])
    )
    assert unionarray.to_list() == [
        0.0,
        1.1,
        2.2,
        [0],
        [100],
        3.3,
        4.4,
        [200],
        [300],
        [400],
    ]

    assert (unionarray + 10).to_list() == [
        10.0,
        11.1,
        12.2,
        [10],
        [110],
        13.3,
        14.4,
        [210],
        [310],
        [410],
    ]
    assert (10 + unionarray).to_list() == [
        10.0,
        11.1,
        12.2,
        [10],
        [110],
        13.3,
        14.4,
        [210],
        [310],
        [410],
    ]

    assert (unionarray + range(0, 100, 10)).to_list() == [
        0.0,
        11.1,
        22.2,
        [30],
        [140],
        53.3,
        64.4,
        [270],
        [380],
        [490],
    ]
    assert (range(0, 100, 10) + unionarray).to_list() == [
        0.0,
        11.1,
        22.2,
        [30],
        [140],
        53.3,
        64.4,
        [270],
        [380],
        [490],
    ]

    assert (unionarray + np.arange(0, 100, 10)).tolist() == [
        0.0,
        11.1,
        22.2,
        [30],
        [140],
        53.3,
        64.4,
        [270],
        [380],
        [490],
    ]
    assert (np.arange(0, 100, 10) + unionarray).tolist() == [
        0.0,
        11.1,
        22.2,
        [30],
        [140],
        53.3,
        64.4,
        [270],
        [380],
        [490],
    ]

    assert (unionarray + ak.highlevel.Array(np.arange(0, 100, 10))).tolist() == [
        0.0,
        11.1,
        22.2,
        [30],
        [140],
        53.3,
        64.4,
        [270],
        [380],
        [490],
    ]
    assert (ak.highlevel.Array(np.arange(0, 100, 10)) + unionarray).tolist() == [
        0.0,
        11.1,
        22.2,
        [30],
        [140],
        53.3,
        64.4,
        [270],
        [380],
        [490],
    ]

    assert (unionarray + unionarray).to_list() == [
        0.0,
        2.2,
        4.4,
        [0],
        [200],
        6.6,
        8.8,
        [400],
        [600],
        [800],
    ]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("from_numpy",)
from awkward._dispatch import high_level_function
from awkward._layout import from_arraylib, wrap_layout


@high_level_function
def from_numpy(
    array, *, regulararray=False, recordarray=True, highlevel=True, behavior=None
):
    """
    Args:
        array (np.ndarray): The NumPy array to convert into an Awkward Array.
            This array can be a np.ma.MaskedArray.
        regulararray (bool): If True and the array is multidimensional,
            the dimensions are represented by nested #ak.contents.RegularArray
            nodes; if False and the array is multidimensional, the dimensions
            are represented by a multivalued #ak.contents.NumpyArray.shape.
            If the array is one-dimensional, this has no effect.
        recordarray (bool): If True and the array is a NumPy structured array
            (dtype.names is not None), the fields are represented by an
            #ak.contents.RecordArray; if False and the array is a structured
            array, the structure is left in the #ak.contents.NumpyArray `format`,
            which some functions do not recognize.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Converts a NumPy array into an Awkward Array.

    The resulting layout can only involve the following #ak.contents.Content types:

    * #ak.contents.NumpyArray
    * #ak.contents.ByteMaskedArray or #ak.contents.UnmaskedArray if the
      `array` is an np.ma.MaskedArray.
    * #ak.contents.RegularArray if `regulararray=True`.
    * #ak.contents.RecordArray if `recordarray=True`.

    See also #ak.to_numpy and #ak.from_cupy.
    """
    return wrap_layout(
        from_arraylib(array, regulararray, recordarray),
        highlevel=highlevel,
        behavior=behavior,
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak

to_list = ak.operations.to_list


def test_numpy_array():
    matrix = np.arange(64).reshape(8, -1)
    layout = ak.contents.NumpyArray(matrix[:, 0])
    assert not layout.is_contiguous

    packed = ak.operations.to_packed(layout, highlevel=False)
    assert to_list(packed) == to_list(layout)
    assert packed.is_contiguous


def test_empty_array():
    layout = ak.contents.EmptyArray()
    assert ak.operations.to_packed(layout, highlevel=False) is layout


def test_indexed_option_array():
    index = ak.index.Index64(np.r_[0, -1, 2, -1, 4])
    content = ak.contents.NumpyArray(np.arange(8))
    layout = ak.contents.IndexedOptionArray(index, content)

    packed = ak.operations.to_packed(layout, highlevel=False)
    assert to_list(layout) == to_list(packed)
    assert isinstance(packed, ak.contents.IndexedOptionArray)
    assert np.asarray(packed.index).tolist() == [0, -1, 1, -1, 2]
    assert len(packed.content) == 3


def test_indexed_array():
    index = ak.index.Index64(np.array([0, 1, 2, 3, 6, 7, 8]))
    content = ak.contents.NumpyArray(np.arange(10))
    layout = ak.contents.IndexedArray(index, content)

    packed = ak.operations.to_packed(layout, highlevel=False)
    assert to_list(packed) == to_list(layout)

    assert isinstance(packed, ak.contents.NumpyArray)
    assert len(packed) == len(index)


def test_list_array():
    content = ak.contents.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    starts = ak.index.Index64(np.array([0, 3, 3, 5, 6]))
    stops = ak.index.Index64(np.array([3, 3, 5, 6, 9]))
    layout = ak.contents.ListArray(starts, stops, content)

    packed = ak.operations.to_packed(layout, highlevel=False)
    assert to_list(packed) == to_list(layout)
    assert isinstance(packed, ak.contents.ListOffsetArray)
    assert packed.offsets[0] == 0


def test_list_offset_array():
    content = ak.contents.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    offsets = ak.index.Index64(np.array([0, 3, 3, 5, 6]))
    layout = ak.contents.ListOffsetArray(offsets, content)

    packed = ak.operations.to_packed(layout, highlevel=False)
    assert to_list(packed) == to_list(layout)
    assert isinstance(packed, ak.contents.ListOffsetArray)
    assert packed.offsets[0] == 0
    assert len(packed.content) == packed.offsets[-1]


def test_unmasked_array():
    content = ak.contents.NumpyArray(
        np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9])
    )
    layout = ak.contents.UnmaskedArray(content)
    packed = ak.operations.to_packed(layout, highlevel=False)
    assert to_list(packed) == to_list(layout)


def test_union_array():
    a = ak.contents.NumpyArray(np.arange(4))
    b = ak.contents.NumpyArray(np.arange(4) + 4)
    c = ak.contents.RegularArray(ak.contents.NumpyArray(np.arange(12)), 3)
    layout = ak.contents.UnionArray.simplified(
        ak.index.Index8([1, 1, 2, 2, 0, 0]),
        ak.index.Index64([0, 1, 0, 1, 0, 1]),
        [a, b, c],
    )
    packed = ak.operations.to_packed(layout, highlevel=False)
    assert to_list(packed) == to_list(layout)
    # Check that it merges like contents
    assert len(packed.contents) == 2
    index_0 = np.asarray(packed.index)[np.asarray(packed.tags) == 0]
    assert index_0.tolist() == [0, 1, 2, 3]


def test_record_array():
    a = ak.contents.NumpyArray(np.arange(10))
    b = ak.contents.NumpyArray(np.arange(10) * 2 + 4)
    layout = ak.contents.RecordArray([a, b], None, 5)
    packed = ak.operations.to_packed(layout, highlevel=False)
    assert to_list(packed) == to_list(layout)
    assert len(packed.contents[0]) == 5
    assert len(packed.contents[1]) == 5


def test_regular_array():
    content = ak.contents.NumpyArray(np.arange(10))
    layout = ak.contents.RegularArray(content, 3)
    packed = ak.operations.to_packed(layout, highlevel=False)
    assert to_list(packed) == to_list(layout)
    assert len(packed.content) == 9
    assert packed.size == layout.size


def test_bit_masked_aray():
    mask = ak.index.IndexU8(np.array([0b10101010]))
    content = ak.contents.NumpyArray(np.arange(16))
    layout = ak.contents.BitMaskedArray(mask, content, False, 8, False)
    packed = ak.operations.to_packed(layout, highlevel=False)
    assert to_list(packed) == to_list(layout)
    assert len(packed.content) == 8


def test_byte_masked_array():
    mask = ak.index.Index8(np.array([1, 0, 1, 0, 1, 0, 1, 0]))
    content = ak.contents.NumpyArray(np.arange(16))
    layout = ak.contents.ByteMaskedArray(
        mask,
        content,
        False,
    )
    packed = ak.operations.to_packed(layout, highlevel=False)
    assert to_list(packed) == to_list(layout)
    assert len(packed.content) == 8


def test_record():
    a = ak.contents.NumpyArray(np.arange(10))
    b = ak.contents.NumpyArray(np.arange(10) * 2 + 4)
    layout = ak.contents.RecordArray([a, b], None, 5)
    record = layout[4]
    packed = ak.operations.to_packed(record, highlevel=False)
    assert to_list(packed) == to_list(record)
    assert len(packed.array) == 1
    assert to_list(packed.array) == [(4, 12)]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
__all__ = ("values_astype",)
import awkward as ak
from awkward._dispatch import high_level_function
from awkward._layout import wrap_layout
from awkward._nplikes.numpylike import NumpyMetadata

np = NumpyMetadata.instance()


@high_level_function
def values_astype(array, to, *, including_unknown=False, highlevel=True, behavior=None):
    """
    Args:
        array: Array-like data (anything #ak.to_layout recognizes).
        to (dtype or dtype specifier): Type to convert the numbers into.
        including_unknown (bool): If True, the `unknown` type is considered
            a value type and is converted to the specified dtype; if False,
            `unknown` will remain `unknown`.
        highlevel (bool): If True, return an #ak.Array; otherwise, return
            a low-level #ak.contents.Content subclass.
        behavior (None or dict): Custom #ak.behavior for the output array, if
            high-level.

    Converts all numbers in the array to a new type, leaving the structure
    untouched.

    For example,

        >>> array = ak.Array([1.1, 2.2, 3.3, 4.4, 5.5])
        >>> ak.values_astype(array, np.int32)
        <Array [1, 2, 3, 4, 5] type='5 * int32'>

    and

        >>> array = ak.Array([[1.1, 2.2, 3.3], [], [4.4, 5.5]])
        >>> ak.values_astype(array, np.int32)
        <Array [[1, 2, 3], [], [4, 5]] type='3 * var * int32'>

    Note, when converting values to a `np.datetime64` type that is unitless, a
    default '[us]' unit is assumed - until further specified as numpy dtypes.

    For example,

        >>> array = ak.Array([1567416600000])
        >>> ak.values_astype(array, "datetime64[ms]")
        <Array [2019-09-02T09:30:00.000] type='1 * datetime64[ms]'>

    or

        >>> array = ak.Array([1567416600000])
        >>> ak.values_astype(array, np.dtype("M8[ms]"))
        <Array [2019-09-02T09:30:00.000] type='1 * datetime64[ms]['>

    See also #ak.strings_astype.
    """
    # Dispatch
    yield (array,)

    # Implementation
    return _impl(array, to, including_unknown, highlevel, behavior)


def _impl(array, to, including_unknown, highlevel, behavior):
    to_dtype = np.dtype(to)
    to_str = ak.types.numpytype.dtype_to_primitive(to_dtype)
    layout = ak.operations.to_layout(array, allow_record=False, allow_other=False)
    out = ak._do.numbers_to_type(layout, to_str, including_unknown)
    return wrap_layout(out, behavior, highlevel, like=array)



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import pytest  # noqa: F401

import awkward as ak


def test():
    a, b, c = ak.broadcast_fields(
        [{"x": 1, "c": 4}],
        [{"y": 2, "c": 5}],
        [{"z": 3, "c": 6}],
    )
    assert ak.almost_equal(a, [{"x": 1, "c": 4, "y": None, "z": None}])
    assert ak.almost_equal(b, [{"x": None, "c": 5, "y": 2, "z": None}])
    assert ak.almost_equal(c, [{"x": None, "c": 6, "y": None, "z": 3}])


def test_nested():
    a, b, c = ak.broadcast_fields(
        [{"x": 1, "z": {"a": 2, "b": 3}}],
        [{"y": 2}],
        [{"y": 3, "z": {"c": 9}}],
    )
    assert ak.almost_equal(a, [{"x": 1, "z": {"a": 2, "b": 3, "c": None}, "y": None}])
    assert ak.almost_equal(b, [{"x": None, "z": None, "y": 2}])
    assert ak.almost_equal(
        c, [{"x": None, "z": {"a": None, "b": None, "c": 9}, "y": 3}]
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest

import awkward as ak

ROOT = pytest.importorskip("ROOT")

import awkward._connect.cling  # noqa: E402
import awkward._lookup  # noqa: E402

compiler = ROOT.gInterpreter.Declare
cpp17 = hasattr(ROOT.std, "optional")


def debug_compiler(code):
    print(code)
    ROOT.gInterpreter.Declare(code)


def test_NumpyArray():
    v2a = ak.contents.numpyarray.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3]),
        parameters={"some": "stuff", "other": [1, 2, "three"]},
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_NumpyArray_v2a(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};
  out[0] = obj.size();
  out[1] = obj[1];
  out[2] = obj[3];
}}
"""
    )
    out = np.zeros(3, dtype=np.float64)
    ROOT.roottest_NumpyArray_v2a(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [4.0, 1.1, 3.3]


def test_EmptyArray():
    v2a = ak.contents.emptyarray.EmptyArray()

    layout = v2a
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
size_t roottest_EmptyArray_v2a(ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};
  return obj.size();
}}
"""
    )
    assert ROOT.roottest_EmptyArray_v2a(len(layout), lookup.arrayptrs) == 0


def test_NumpyArray_shape():
    v2a = ak.contents.numpyarray.NumpyArray(
        np.arange(2 * 3 * 5, dtype=np.int64).reshape(2, 3, 5)
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_NumpyArray_shape_v2a(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};
  out[0] = obj.size();
  out[1] = obj[0].size();
  out[2] = obj[0][0].size();
  out[3] = obj[0][0][0];
  out[4] = obj[0][0][1];
  out[5] = obj[0][1][0];
  out[6] = obj[0][1][1];
  out[7] = obj[1][0][0];
  out[8] = obj[1][1][1];
}}
"""
    )
    out = np.zeros(9, dtype=np.float64)
    ROOT.roottest_NumpyArray_shape_v2a(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [2.0, 3.0, 5.0, 0.0, 1.0, 5.0, 6.0, 15.0, 21.0]


@pytest.mark.parametrize("flatlist_as_rvec", [False, True])
def test_RegularArray_NumpyArray(flatlist_as_rvec):
    v2a = ak.contents.regulararray.RegularArray(
        ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
        3,
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(
        layout.form, flatlist_as_rvec=flatlist_as_rvec
    )
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_RegularArray_NumpyArray_v2a_{flatlist_as_rvec}(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};
  out[0] = obj.size();
  out[1] = obj[0][0];
  out[2] = obj[0][1];
  out[3] = obj[1][0];
  out[4] = obj[1][1];
  out[5] = obj[1].size();
}}
"""
    )
    out = np.zeros(6, dtype=np.float64)
    getattr(ROOT, f"roottest_RegularArray_NumpyArray_v2a_{flatlist_as_rvec}")(
        out, len(layout), lookup.arrayptrs
    )
    assert out.tolist() == [2.0, 0.0, 1.1, 3.3, 4.4, 3.0]


def test_RegularArray_NumpyArray_v2b():
    v2b = ak.contents.regulararray.RegularArray(
        ak.contents.emptyarray.EmptyArray().to_NumpyArray(np.dtype(np.float64)),
        0,
        zeros_length=10,
    )

    layout = v2b
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_RegularArray_NumpyArray_v2b(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};
  out[0] = obj.size();
  out[1] = obj[0].size();
  out[2] = obj[1].size();
}}
"""
    )
    out = np.zeros(3, dtype=np.float64)
    ROOT.roottest_RegularArray_NumpyArray_v2b(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [10.0, 0.0, 0.0]


@pytest.mark.parametrize("flatlist_as_rvec", [False, True])
def test_ListArray_NumpyArray(flatlist_as_rvec):
    v2a = ak.contents.listarray.ListArray(
        ak.index.Index(np.array([4, 100, 1], np.int64)),
        ak.index.Index(np.array([7, 100, 3, 200], np.int64)),
        ak.contents.numpyarray.NumpyArray(
            np.array([6.6, 4.4, 5.5, 7.7, 1.1, 2.2, 3.3, 8.8])
        ),
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(
        layout.form, flatlist_as_rvec=flatlist_as_rvec
    )
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_ListArray_NumpyArray_v2a_{flatlist_as_rvec}(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};
  out[0] = obj.size();
  out[1] = obj[0].size();
  out[2] = obj[0][0];
  out[3] = obj[0][1];
  out[4] = obj[0][2];
  out[5] = obj[1].size();
  out[6] = obj[2].size();
  out[7] = obj[2][0];
  out[8] = obj[2][1];
}}
"""
    )
    out = np.zeros(9, dtype=np.float64)
    getattr(ROOT, f"roottest_ListArray_NumpyArray_v2a_{flatlist_as_rvec}")(
        out, len(layout), lookup.arrayptrs
    )
    assert out.tolist() == [3.0, 3.0, 1.1, 2.2, 3.3, 0.0, 2.0, 4.4, 5.5]


@pytest.mark.parametrize("flatlist_as_rvec", [False, True])
def test_ListOffsetArray_NumpyArray(flatlist_as_rvec):
    v2a = ak.contents.listoffsetarray.ListOffsetArray(
        ak.index.Index(np.array([1, 4, 4, 6, 7], np.int64)),
        ak.contents.numpyarray.NumpyArray(
            np.array([6.6, 1.1, 2.2, 3.3, 4.4, 5.5, 7.7])
        ),
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(
        layout.form, flatlist_as_rvec=flatlist_as_rvec
    )
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_ListOffsetArray_NumpyArray_{flatlist_as_rvec}(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};
  out[0] = obj.size();
  out[1] = obj[0].size();
  out[2] = obj[0][0];
  out[3] = obj[0][1];
  out[4] = obj[0][2];
  out[5] = obj[1].size();
  out[6] = obj[2].size();
  out[7] = obj[2][0];
  out[8] = obj[2][1];
  out[9] = obj[3].size();
  out[10] = obj[3][0];
}}
"""
    )
    out = np.zeros(11, dtype=np.float64)
    getattr(ROOT, f"roottest_ListOffsetArray_NumpyArray_{flatlist_as_rvec}")(
        out, len(layout), lookup.arrayptrs
    )
    assert out.tolist() == [4.0, 3.0, 1.1, 2.2, 3.3, 0.0, 2.0, 4.4, 5.5, 1.0, 7.7]


def test_RecordArray_NumpyArray():
    v2a = ak.contents.recordarray.RecordArray(
        [
            ak.contents.numpyarray.NumpyArray(np.array([0, 1, 2, 3, 4], np.int64)),
            ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
        ],
        ["x", "y"],
        parameters={"__record__": "Something"},
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_RecordArray_NumpyArray_v2a(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};
  out[0] = obj.size();
  auto rec1 = obj[1];
  auto rec4 = obj[4];
  out[1] = rec1.x();
  out[2] = rec1.y();
  out[3] = rec4.x();
  out[4] = rec4.y();
}}
"""
    )
    out = np.zeros(5, dtype=np.float64)
    ROOT.roottest_RecordArray_NumpyArray_v2a(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [5.0, 1, 1.1, 4, 4.4]

    v2b = ak.contents.recordarray.RecordArray(
        [
            ak.contents.numpyarray.NumpyArray(np.array([0, 1, 2, 3, 4], np.int64)),
            ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
        ],
        None,
    )

    layout = v2b
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_RecordArray_NumpyArray_v2b(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};
  out[0] = obj.size();
  auto rec1 = obj[1];
  auto rec4 = obj[4];
  out[1] = rec1.slot0();
  out[2] = rec1.slot1();
  out[3] = rec4.slot0();
  out[4] = rec4.slot1();
}}
"""
    )
    out = np.zeros(5, dtype=np.float64)
    ROOT.roottest_RecordArray_NumpyArray_v2b(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [5.0, 1, 1.1, 4, 4.4]

    v2c = ak.contents.recordarray.RecordArray([], [], 10)

    layout = v2c
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_RecordArray_NumpyArray_v2c(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};
  out[0] = obj.size();
  obj[5];
}}
"""
    )
    out = np.zeros(1, dtype=np.float64)
    ROOT.roottest_RecordArray_NumpyArray_v2c(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [10.0]

    v2d = ak.contents.recordarray.RecordArray([], None, 10)

    layout = v2d
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_RecordArray_NumpyArray_v2d(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};
  out[0] = obj.size();
  obj[5];
}}
"""
    )
    out = np.zeros(1, dtype=np.float64)
    ROOT.roottest_RecordArray_NumpyArray_v2d(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [10.0]


@pytest.mark.skipif(not cpp17, reason="ROOT was compiled without C++17 support")
def test_IndexedArray_NumpyArray():
    v2a = ak.contents.indexedarray.IndexedArray(
        ak.index.Index(np.array([2, 2, 0, 1, 4, 5, 4], np.int64)),
        ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_IndexedArray_NumpyArray_v2a(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};
  out[0] = obj.size();
  out[1] = obj[0];
  out[2] = obj[1];
  out[3] = obj[2];
  out[4] = obj[3];
  out[5] = obj[4];
  out[6] = obj[5];
  out[7] = obj[6];
}}
"""
    )
    out = np.zeros(8, dtype=np.float64)
    ROOT.roottest_IndexedArray_NumpyArray_v2a(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [7.0, 2.2, 2.2, 0.0, 1.1, 4.4, 5.5, 4.4]


def test_IndexedOptionArray_NumpyArray():
    v2a = ak.contents.indexedoptionarray.IndexedOptionArray(
        ak.index.Index(np.array([2, 2, -1, 1, -1, 5, 4], np.int64)),
        ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_IndexedOptionArray_NumpyArray_v2a(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};
  out[0] = obj.size();
  out[1] = obj[0].has_value() ? obj[0].value() : 999.0;
  out[2] = obj[1].has_value() ? obj[1].value() : 999.0;
  out[3] = obj[2].has_value() ? obj[2].value() : 999.0;
  out[4] = obj[3].has_value() ? obj[3].value() : 999.0;
  out[5] = obj[4].has_value() ? obj[4].value() : 999.0;
  out[6] = obj[5].has_value() ? obj[5].value() : 999.0;
  out[7] = obj[6].has_value() ? obj[6].value() : 999.0;
}}
"""
    )
    out = np.zeros(8, dtype=np.float64)
    ROOT.roottest_IndexedOptionArray_NumpyArray_v2a(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [7.0, 2.2, 2.2, 999.0, 1.1, 999.0, 5.5, 4.4]


@pytest.mark.skipif(not cpp17, reason="ROOT was compiled without C++17 support")
def test_ByteMaskedArray_NumpyArray():
    v2a = ak.contents.bytemaskedarray.ByteMaskedArray(
        ak.index.Index(np.array([1, 0, 1, 0, 1], np.int8)),
        ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])),
        valid_when=True,
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_ByteMaskedArray_NumpyArray_v2a(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};
  out[0] = obj.size();
  out[1] = obj[0].has_value() ? obj[0].value() : 999.0;
  out[2] = obj[1].has_value() ? obj[1].value() : 999.0;
  out[3] = obj[2].has_value() ? obj[2].value() : 999.0;
  out[4] = obj[3].has_value() ? obj[3].value() : 999.0;
  out[5] = obj[4].has_value() ? obj[4].value() : 999.0;
}}
"""
    )
    out = np.zeros(6, dtype=np.float64)
    ROOT.roottest_ByteMaskedArray_NumpyArray_v2a(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [5.0, 1.1, 999.0, 3.3, 999.0, 5.5]

    v2b = ak.contents.bytemaskedarray.ByteMaskedArray(
        ak.index.Index(np.array([0, 1, 0, 1, 0], np.int8)),
        ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])),
        valid_when=False,
    )

    layout = v2b
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_ByteMaskedArray_NumpyArray_v2b(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};
  out[0] = obj.size();
  out[1] = obj[0].has_value() ? obj[0].value() : 999.0;
  out[2] = obj[1].has_value() ? obj[1].value() : 999.0;
  out[3] = obj[2].has_value() ? obj[2].value() : 999.0;
  out[4] = obj[3].has_value() ? obj[3].value() : 999.0;
  out[5] = obj[4].has_value() ? obj[4].value() : 999.0;
}}
"""
    )
    out = np.zeros(6, dtype=np.float64)
    ROOT.roottest_ByteMaskedArray_NumpyArray_v2b(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [5.0, 1.1, 999.0, 3.3, 999.0, 5.5]


@pytest.mark.skipif(not cpp17, reason="ROOT was compiled without C++17 support")
def test_BitMaskedArray_NumpyArray():
    v2a = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                        1,
                    ],
                    np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=True,
        length=13,
        lsb_order=False,
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_BitMaskedArray_NumpyArray_v2a(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};
  out[0] = obj.size();
  out[1] = obj[0].has_value() ? obj[0].value() : 999.0;
  out[2] = obj[1].has_value() ? obj[1].value() : 999.0;
  out[3] = obj[2].has_value() ? obj[2].value() : 999.0;
  out[4] = obj[3].has_value() ? obj[3].value() : 999.0;
  out[5] = obj[4].has_value() ? obj[4].value() : 999.0;
  out[6] = obj[5].has_value() ? obj[5].value() : 999.0;
  out[7] = obj[6].has_value() ? obj[6].value() : 999.0;
  out[8] = obj[7].has_value() ? obj[7].value() : 999.0;
  out[9] = obj[8].has_value() ? obj[8].value() : 999.0;
  out[10] = obj[9].has_value() ? obj[9].value() : 999.0;
  out[11] = obj[10].has_value() ? obj[10].value() : 999.0;
  out[12] = obj[11].has_value() ? obj[11].value() : 999.0;
  out[13] = obj[12].has_value() ? obj[12].value() : 999.0;
}}
"""
    )
    out = np.zeros(14, dtype=np.float64)
    ROOT.roottest_BitMaskedArray_NumpyArray_v2a(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [
        13.0,
        0.0,
        1.0,
        2.0,
        3.0,
        999.0,
        999.0,
        999.0,
        999.0,
        1.1,
        999.0,
        3.3,
        999.0,
        5.5,
    ]

    v2b = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        1,
                        0,
                        1,
                        0,
                    ],
                    np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=False,
        length=13,
        lsb_order=False,
    )

    layout = v2b
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_BitMaskedArray_NumpyArray_v2b(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};
  out[0] = obj.size();
  out[1] = obj[0].has_value() ? obj[0].value() : 999.0;
  out[2] = obj[1].has_value() ? obj[1].value() : 999.0;
  out[3] = obj[2].has_value() ? obj[2].value() : 999.0;
  out[4] = obj[3].has_value() ? obj[3].value() : 999.0;
  out[5] = obj[4].has_value() ? obj[4].value() : 999.0;
  out[6] = obj[5].has_value() ? obj[5].value() : 999.0;
  out[7] = obj[6].has_value() ? obj[6].value() : 999.0;
  out[8] = obj[7].has_value() ? obj[7].value() : 999.0;
  out[9] = obj[8].has_value() ? obj[8].value() : 999.0;
  out[10] = obj[9].has_value() ? obj[9].value() : 999.0;
  out[11] = obj[10].has_value() ? obj[10].value() : 999.0;
  out[12] = obj[11].has_value() ? obj[11].value() : 999.0;
  out[13] = obj[12].has_value() ? obj[12].value() : 999.0;
}}
"""
    )
    out = np.zeros(14, dtype=np.float64)
    ROOT.roottest_BitMaskedArray_NumpyArray_v2b(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [
        13.0,
        0.0,
        1.0,
        2.0,
        3.0,
        999.0,
        999.0,
        999.0,
        999.0,
        1.1,
        999.0,
        3.3,
        999.0,
        5.5,
    ]

    v2c = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                        1,
                    ],
                    np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=True,
        length=13,
        lsb_order=True,
    )

    layout = v2c
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_BitMaskedArray_NumpyArray_v2c(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};
  out[0] = obj.size();
  out[1] = obj[0].has_value() ? obj[0].value() : 999.0;
  out[2] = obj[1].has_value() ? obj[1].value() : 999.0;
  out[3] = obj[2].has_value() ? obj[2].value() : 999.0;
  out[4] = obj[3].has_value() ? obj[3].value() : 999.0;
  out[5] = obj[4].has_value() ? obj[4].value() : 999.0;
  out[6] = obj[5].has_value() ? obj[5].value() : 999.0;
  out[7] = obj[6].has_value() ? obj[6].value() : 999.0;
  out[8] = obj[7].has_value() ? obj[7].value() : 999.0;
  out[9] = obj[8].has_value() ? obj[8].value() : 999.0;
  out[10] = obj[9].has_value() ? obj[9].value() : 999.0;
  out[11] = obj[10].has_value() ? obj[10].value() : 999.0;
  out[12] = obj[11].has_value() ? obj[11].value() : 999.0;
  out[13] = obj[12].has_value() ? obj[12].value() : 999.0;
}}
"""
    )
    out = np.zeros(14, dtype=np.float64)
    ROOT.roottest_BitMaskedArray_NumpyArray_v2c(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [
        13.0,
        0.0,
        1.0,
        2.0,
        3.0,
        999.0,
        999.0,
        999.0,
        999.0,
        1.1,
        999.0,
        3.3,
        999.0,
        5.5,
    ]

    v2d = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                    ],
                    np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=False,
        length=13,
        lsb_order=True,
    )

    layout = v2d
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_BitMaskedArray_NumpyArray_v2d(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};
  out[0] = obj.size();
  out[1] = obj[0].has_value() ? obj[0].value() : 999.0;
  out[2] = obj[1].has_value() ? obj[1].value() : 999.0;
  out[3] = obj[2].has_value() ? obj[2].value() : 999.0;
  out[4] = obj[3].has_value() ? obj[3].value() : 999.0;
  out[5] = obj[4].has_value() ? obj[4].value() : 999.0;
  out[6] = obj[5].has_value() ? obj[5].value() : 999.0;
  out[7] = obj[6].has_value() ? obj[6].value() : 999.0;
  out[8] = obj[7].has_value() ? obj[7].value() : 999.0;
  out[9] = obj[8].has_value() ? obj[8].value() : 999.0;
  out[10] = obj[9].has_value() ? obj[9].value() : 999.0;
  out[11] = obj[10].has_value() ? obj[10].value() : 999.0;
  out[12] = obj[11].has_value() ? obj[11].value() : 999.0;
  out[13] = obj[12].has_value() ? obj[12].value() : 999.0;
}}
"""
    )
    out = np.zeros(14, dtype=np.float64)
    ROOT.roottest_BitMaskedArray_NumpyArray_v2d(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [
        13.0,
        0.0,
        1.0,
        2.0,
        3.0,
        999.0,
        999.0,
        999.0,
        999.0,
        1.1,
        999.0,
        3.3,
        999.0,
        5.5,
    ]


@pytest.mark.skipif(not cpp17, reason="ROOT was compiled without C++17 support")
def test_UnmaskedArray_NumpyArray():
    v2a = ak.contents.unmaskedarray.UnmaskedArray(
        ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3]))
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_UnmaskedArray_NumpyArray_v2a(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};
  out[0] = obj.size();
  out[1] = obj[1].has_value() ? obj[1].value() : 999.0;
  out[2] = obj[3].has_value() ? obj[3].value() : 999.0;
}}
"""
    )
    out = np.zeros(3, dtype=np.float64)
    ROOT.roottest_UnmaskedArray_NumpyArray_v2a(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [4.0, 1.1, 3.3]


def test_UnionArray_NumpyArray():
    v2a = ak.contents.unionarray.UnionArray(
        ak.index.Index(np.array([1, 1, 0, 0, 1, 0, 1], np.int8)),
        ak.index.Index(np.array([4, 3, 0, 1, 2, 2, 4, 100], np.int64)),
        [
            ak.from_iter(["1", "2", "3"], highlevel=False),
            ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5])),
        ],
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_UnionArray_NumpyArray_v2a(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};
  out[0] = obj.size();
  out[1] = obj[0].index();
  out[2] = obj[1].index();
  out[3] = obj[2].index();
  out[4] = obj[3].index();
  out[5] = obj[4].index();
  out[6] = obj[5].index();
  out[7] = obj[6].index();
  out[8] = std::get<double>(obj[0]);
  out[9] = std::get<double>(obj[1]);
  out[10] = std::atoi(std::get<std::string>(obj[2]).c_str());
  out[11] = std::atoi(std::get<std::string>(obj[3]).c_str());
  out[12] = std::get<double>(obj[4]);
  out[13] = std::atoi(std::get<std::string>(obj[5]).c_str());
  out[14] = std::get<double>(obj[6]);
}}
"""
    )
    out = np.zeros(15, dtype=np.float64)
    ROOT.roottest_UnionArray_NumpyArray_v2a(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [7, 1, 1, 0, 0, 1, 0, 1, 5.5, 4.4, 1, 2, 3.3, 3, 5.5]


@pytest.mark.parametrize("flatlist_as_rvec", [False, True])
def test_nested_NumpyArray(flatlist_as_rvec):
    v2a = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 5], dtype=np.int64)),
        ak.contents.numpyarray.NumpyArray(
            np.array([999.0, 0.0, 1.1, 2.2, 3.3]),
            parameters={"some": "stuff", "other": [1, 2, "three"]},
        ),
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(
        layout.form, flatlist_as_rvec=flatlist_as_rvec
    )
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_nested_NumpyArray_v2a_{flatlist_as_rvec}(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()}[1];
  out[0] = obj.size();
  out[1] = obj[1];
  out[2] = obj[3];
}}
"""
    )
    out = np.zeros(3, dtype=np.float64)
    getattr(ROOT, f"roottest_nested_NumpyArray_v2a_{flatlist_as_rvec}")(
        out, len(layout), lookup.arrayptrs
    )
    assert out.tolist() == [4.0, 1.1, 3.3]


def test_nested_NumpyArray_shape():
    data = np.full((3, 3, 5), 999, dtype=np.int64)
    data[1:3] = np.arange(2 * 3 * 5, dtype=np.int64).reshape(2, 3, 5)

    v2a = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 3], dtype=np.int64)),
        ak.contents.numpyarray.NumpyArray(data),
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_nested_NumpyArray_shape_v2a(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()}[1];
  out[0] = obj.size();
  out[1] = obj[0].size();
  out[2] = obj[0][0].size();
  out[3] = obj[0][0][0];
  out[4] = obj[0][0][1];
  out[5] = obj[0][1][0];
  out[6] = obj[0][1][1];
  out[7] = obj[1][0][0];
  out[8] = obj[1][1][1];
}}
"""
    )
    out = np.zeros(9, dtype=np.float64)
    ROOT.roottest_nested_NumpyArray_shape_v2a(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [2.0, 3.0, 5.0, 0.0, 1.0, 5.0, 6.0, 15.0, 21.0]


@pytest.mark.parametrize("flatlist_as_rvec", [False, True])
def test_nested_RegularArray_NumpyArray(flatlist_as_rvec):
    v2a = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 3], dtype=np.int64)),
        ak.contents.regulararray.RegularArray(
            ak.contents.numpyarray.NumpyArray(
                np.array([999, 999, 999, 0.0, 1.1, 2.2, 3.3, 4.4, 5.5])
            ),
            3,
        ),
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(
        layout.form, flatlist_as_rvec=flatlist_as_rvec
    )
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_nested_RegularArray_NumpyArray_v2a_{flatlist_as_rvec}(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()}[1];
  out[0] = obj.size();
  out[1] = obj[0][0];
  out[2] = obj[0][1];
  out[3] = obj[1][0];
  out[4] = obj[1][1];
  out[5] = obj[1].size();
}}
"""
    )
    out = np.zeros(6, dtype=np.float64)
    getattr(ROOT, f"roottest_nested_RegularArray_NumpyArray_v2a_{flatlist_as_rvec}")(
        out, len(layout), lookup.arrayptrs
    )
    assert out.tolist() == [2.0, 0.0, 1.1, 3.3, 4.4, 3.0]

    v2b = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 11], dtype=np.int64)),
        ak.contents.regulararray.RegularArray(
            ak.contents.emptyarray.EmptyArray().to_NumpyArray(np.dtype(np.float64)),
            0,
            zeros_length=11,
        ),
    )

    layout = v2b
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_nested_RegularArray_NumpyArray_v2b{flatlist_as_rvec}(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()}[1];
  out[0] = obj.size();
  out[1] = obj[0].size();
  out[2] = obj[1].size();
}}
"""
    )
    out = np.zeros(3, dtype=np.float64)
    getattr(ROOT, f"roottest_nested_RegularArray_NumpyArray_v2b{flatlist_as_rvec}")(
        out, len(layout), lookup.arrayptrs
    )
    assert out.tolist() == [10.0, 0.0, 0.0]


@pytest.mark.parametrize("flatlist_as_rvec", [False, True])
def test_nested_ListArray_NumpyArray(flatlist_as_rvec):
    v2a = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 4], dtype=np.int64)),
        ak.contents.listarray.ListArray(
            ak.index.Index(np.array([999, 4, 100, 1], np.int64)),
            ak.index.Index(np.array([999, 7, 100, 3, 200], np.int64)),
            ak.contents.numpyarray.NumpyArray(
                np.array([6.6, 4.4, 5.5, 7.7, 1.1, 2.2, 3.3, 8.8])
            ),
        ),
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(
        layout.form, flatlist_as_rvec=flatlist_as_rvec
    )
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_nested_ListArray_NumpyArray_v2a_{flatlist_as_rvec}(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()}[1];
  out[0] = obj.size();
  out[1] = obj[0].size();
  out[2] = obj[0][0];
  out[3] = obj[0][1];
  out[4] = obj[0][2];
  out[5] = obj[1].size();
  out[6] = obj[2].size();
  out[7] = obj[2][0];
  out[8] = obj[2][1];
}}
"""
    )
    out = np.zeros(9, dtype=np.float64)
    getattr(ROOT, f"roottest_nested_ListArray_NumpyArray_v2a_{flatlist_as_rvec}")(
        out, len(layout), lookup.arrayptrs
    )
    assert out.tolist() == [3.0, 3.0, 1.1, 2.2, 3.3, 0.0, 2.0, 4.4, 5.5]


@pytest.mark.parametrize("flatlist_as_rvec", [False, True])
def test_nested_ListOffsetArray_NumpyArray(flatlist_as_rvec):
    v2a = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 5], dtype=np.int64)),
        ak.contents.listoffsetarray.ListOffsetArray(
            ak.index.Index(np.array([1, 1, 4, 4, 6, 7], np.int64)),
            ak.contents.numpyarray.NumpyArray(
                np.array([6.6, 1.1, 2.2, 3.3, 4.4, 5.5, 7.7])
            ),
        ),
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(
        layout.form, flatlist_as_rvec=flatlist_as_rvec
    )
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_nested_ListOffsetArray_NumpyArray_{flatlist_as_rvec}(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()}[1];
  out[0] = obj.size();
  out[1] = obj[0].size();
  out[2] = obj[0][0];
  out[3] = obj[0][1];
  out[4] = obj[0][2];
  out[5] = obj[1].size();
  out[6] = obj[2].size();
  out[7] = obj[2][0];
  out[8] = obj[2][1];
  out[9] = obj[3].size();
  out[10] = obj[3][0];
}}
"""
    )
    out = np.zeros(11, dtype=np.float64)
    getattr(ROOT, f"roottest_nested_ListOffsetArray_NumpyArray_{flatlist_as_rvec}")(
        out, len(layout), lookup.arrayptrs
    )
    assert out.tolist() == [4.0, 3.0, 1.1, 2.2, 3.3, 0.0, 2.0, 4.4, 5.5, 1.0, 7.7]


def test_nested_RecordArray_NumpyArray():
    v2a = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 6], dtype=np.int64)),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([999, 0, 1, 2, 3, 4], np.int64)
                ),
                ak.contents.numpyarray.NumpyArray(
                    np.array([999, 0.0, 1.1, 2.2, 3.3, 4.4, 5.5])
                ),
            ],
            ["x", "y"],
            parameters={"__record__": "Something"},
        ),
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_nested_RecordArray_NumpyArray_v2a(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()}[1];
  out[0] = obj.size();
  auto rec1 = obj[1];
  auto rec4 = obj[4];
  out[1] = rec1.x();
  out[2] = rec1.y();
  out[3] = rec4.x();
  out[4] = rec4.y();
}}
"""
    )
    out = np.zeros(5, dtype=np.float64)
    ROOT.roottest_nested_RecordArray_NumpyArray_v2a(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [5.0, 1, 1.1, 4, 4.4]

    v2b = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 6], dtype=np.int64)),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([999, 0, 1, 2, 3, 4], np.int64)
                ),
                ak.contents.numpyarray.NumpyArray(
                    np.array([999, 0.0, 1.1, 2.2, 3.3, 4.4, 5.5])
                ),
            ],
            None,
        ),
    )

    layout = v2b
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_nested_RecordArray_NumpyArray_v2b(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()}[1];
  out[0] = obj.size();
  auto rec1 = obj[1];
  auto rec4 = obj[4];
  out[1] = rec1.slot0();
  out[2] = rec1.slot1();
  out[3] = rec4.slot0();
  out[4] = rec4.slot1();
}}
"""
    )
    out = np.zeros(5, dtype=np.float64)
    ROOT.roottest_nested_RecordArray_NumpyArray_v2b(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [5.0, 1, 1.1, 4, 4.4]

    v2c = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 11], dtype=np.int64)),
        ak.contents.recordarray.RecordArray([], [], 11),
    )

    layout = v2c
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_nested_RecordArray_NumpyArray_v2c(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()}[1];
  out[0] = obj.size();
  obj[5];
}}
"""
    )
    out = np.zeros(1, dtype=np.float64)
    ROOT.roottest_nested_RecordArray_NumpyArray_v2c(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [10.0]

    v2d = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 11], dtype=np.int64)),
        ak.contents.recordarray.RecordArray([], None, 11),
    )

    layout = v2d
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_nested_RecordArray_NumpyArray_v2d(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()}[1];
  out[0] = obj.size();
  obj[5];
}}
"""
    )
    out = np.zeros(1, dtype=np.float64)
    ROOT.roottest_nested_RecordArray_NumpyArray_v2d(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [10.0]


def test_nested_IndexedArray_NumpyArray():
    v2a = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 8], dtype=np.int64)),
        ak.contents.indexedarray.IndexedArray(
            ak.index.Index(np.array([999, 2, 2, 0, 1, 4, 5, 4], np.int64)),
            ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
        ),
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_nested_IndexedArray_NumpyArray_v2a(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()}[1];
  out[0] = obj.size();
  out[1] = obj[0];
  out[2] = obj[1];
  out[3] = obj[2];
  out[4] = obj[3];
  out[5] = obj[4];
  out[6] = obj[5];
  out[7] = obj[6];
}}
"""
    )
    out = np.zeros(8, dtype=np.float64)
    ROOT.roottest_nested_IndexedArray_NumpyArray_v2a(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [7.0, 2.2, 2.2, 0.0, 1.1, 4.4, 5.5, 4.4]


def test_nested_IndexedOptionArray_NumpyArray():
    v2a = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 8], dtype=np.int64)),
        ak.contents.indexedoptionarray.IndexedOptionArray(
            ak.index.Index(np.array([999, 2, 2, -1, 1, -1, 5, 4], np.int64)),
            ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
        ),
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_nested_IndexedOptionArray_NumpyArray_v2a(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()}[1];
  out[0] = obj.size();
  out[1] = obj[0].has_value() ? obj[0].value() : 999.0;
  out[2] = obj[1].has_value() ? obj[1].value() : 999.0;
  out[3] = obj[2].has_value() ? obj[2].value() : 999.0;
  out[4] = obj[3].has_value() ? obj[3].value() : 999.0;
  out[5] = obj[4].has_value() ? obj[4].value() : 999.0;
  out[6] = obj[5].has_value() ? obj[5].value() : 999.0;
  out[7] = obj[6].has_value() ? obj[6].value() : 999.0;
}}
"""
    )
    out = np.zeros(8, dtype=np.float64)
    ROOT.roottest_nested_IndexedOptionArray_NumpyArray_v2a(
        out, len(layout), lookup.arrayptrs
    )
    assert out.tolist() == [7.0, 2.2, 2.2, 999.0, 1.1, 999.0, 5.5, 4.4]


def test_nested_ByteMaskedArray_NumpyArray():
    v2a = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 6], dtype=np.int64)),
        ak.contents.bytemaskedarray.ByteMaskedArray(
            ak.index.Index(np.array([123, 1, 0, 1, 0, 1], np.int8)),
            ak.contents.numpyarray.NumpyArray(
                np.array([999, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
            ),
            valid_when=True,
        ),
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_nested_ByteMaskedArray_NumpyArray_v2a(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()}[1];
  out[0] = obj.size();
  out[1] = obj[0].has_value() ? obj[0].value() : 999.0;
  out[2] = obj[1].has_value() ? obj[1].value() : 999.0;
  out[3] = obj[2].has_value() ? obj[2].value() : 999.0;
  out[4] = obj[3].has_value() ? obj[3].value() : 999.0;
  out[5] = obj[4].has_value() ? obj[4].value() : 999.0;
}}
"""
    )
    out = np.zeros(6, dtype=np.float64)
    ROOT.roottest_nested_ByteMaskedArray_NumpyArray_v2a(
        out, len(layout), lookup.arrayptrs
    )
    assert out.tolist() == [5.0, 1.1, 999.0, 3.3, 999.0, 5.5]

    v2b = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 6], dtype=np.int64)),
        ak.contents.bytemaskedarray.ByteMaskedArray(
            ak.index.Index(np.array([123, 0, 1, 0, 1, 0], np.int8)),
            ak.contents.numpyarray.NumpyArray(
                np.array([999, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
            ),
            valid_when=False,
        ),
    )

    layout = v2b
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_nested_ByteMaskedArray_NumpyArray_v2b(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()}[1];
  out[0] = obj.size();
  out[1] = obj[0].has_value() ? obj[0].value() : 999.0;
  out[2] = obj[1].has_value() ? obj[1].value() : 999.0;
  out[3] = obj[2].has_value() ? obj[2].value() : 999.0;
  out[4] = obj[3].has_value() ? obj[3].value() : 999.0;
  out[5] = obj[4].has_value() ? obj[4].value() : 999.0;
}}
"""
    )
    out = np.zeros(6, dtype=np.float64)
    ROOT.roottest_nested_ByteMaskedArray_NumpyArray_v2b(
        out, len(layout), lookup.arrayptrs
    )
    assert out.tolist() == [5.0, 1.1, 999.0, 3.3, 999.0, 5.5]


def test_nested_BitMaskedArray_NumpyArray():
    v2a = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 14], dtype=np.int64)),
        ak.contents.bitmaskedarray.BitMaskedArray(
            ak.index.Index(
                np.packbits(
                    np.array(
                        [
                            0,
                            1,
                            1,
                            1,
                            1,
                            0,
                            0,
                            0,
                            0,
                            1,
                            0,
                            1,
                            0,
                            1,
                        ],
                        np.uint8,
                    )
                )
            ),
            ak.contents.numpyarray.NumpyArray(
                np.array(
                    [
                        999,
                        0.0,
                        1.0,
                        2.0,
                        3.0,
                        4.0,
                        5.0,
                        6.0,
                        7.0,
                        1.1,
                        2.2,
                        3.3,
                        4.4,
                        5.5,
                        6.6,
                    ]
                )
            ),
            valid_when=True,
            length=14,
            lsb_order=False,
        ),
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_nested_BitMaskedArray_NumpyArray_v2a(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()}[1];
  out[0] = obj.size();
  out[1] = obj[0].has_value() ? obj[0].value() : 999.0;
  out[2] = obj[1].has_value() ? obj[1].value() : 999.0;
  out[3] = obj[2].has_value() ? obj[2].value() : 999.0;
  out[4] = obj[3].has_value() ? obj[3].value() : 999.0;
  out[5] = obj[4].has_value() ? obj[4].value() : 999.0;
  out[6] = obj[5].has_value() ? obj[5].value() : 999.0;
  out[7] = obj[6].has_value() ? obj[6].value() : 999.0;
  out[8] = obj[7].has_value() ? obj[7].value() : 999.0;
  out[9] = obj[8].has_value() ? obj[8].value() : 999.0;
  out[10] = obj[9].has_value() ? obj[9].value() : 999.0;
  out[11] = obj[10].has_value() ? obj[10].value() : 999.0;
  out[12] = obj[11].has_value() ? obj[11].value() : 999.0;
  out[13] = obj[12].has_value() ? obj[12].value() : 999.0;
}}
"""
    )
    out = np.zeros(14, dtype=np.float64)
    ROOT.roottest_nested_BitMaskedArray_NumpyArray_v2a(
        out, len(layout), lookup.arrayptrs
    )
    assert out.tolist() == [
        13.0,
        0.0,
        1.0,
        2.0,
        3.0,
        999.0,
        999.0,
        999.0,
        999.0,
        1.1,
        999.0,
        3.3,
        999.0,
        5.5,
    ]

    v2b = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 14], dtype=np.int64)),
        ak.contents.bitmaskedarray.BitMaskedArray(
            ak.index.Index(
                np.packbits(
                    np.array(
                        [
                            0,
                            0,
                            0,
                            0,
                            0,
                            1,
                            1,
                            1,
                            1,
                            0,
                            1,
                            0,
                            1,
                            0,
                        ],
                        np.uint8,
                    )
                )
            ),
            ak.contents.numpyarray.NumpyArray(
                np.array(
                    [
                        999,
                        0.0,
                        1.0,
                        2.0,
                        3.0,
                        4.0,
                        5.0,
                        6.0,
                        7.0,
                        1.1,
                        2.2,
                        3.3,
                        4.4,
                        5.5,
                        6.6,
                    ]
                )
            ),
            valid_when=False,
            length=14,
            lsb_order=False,
        ),
    )

    layout = v2b
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_nested_BitMaskedArray_NumpyArray_v2b(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()}[1];
  out[0] = obj.size();
  out[1] = obj[0].has_value() ? obj[0].value() : 999.0;
  out[2] = obj[1].has_value() ? obj[1].value() : 999.0;
  out[3] = obj[2].has_value() ? obj[2].value() : 999.0;
  out[4] = obj[3].has_value() ? obj[3].value() : 999.0;
  out[5] = obj[4].has_value() ? obj[4].value() : 999.0;
  out[6] = obj[5].has_value() ? obj[5].value() : 999.0;
  out[7] = obj[6].has_value() ? obj[6].value() : 999.0;
  out[8] = obj[7].has_value() ? obj[7].value() : 999.0;
  out[9] = obj[8].has_value() ? obj[8].value() : 999.0;
  out[10] = obj[9].has_value() ? obj[9].value() : 999.0;
  out[11] = obj[10].has_value() ? obj[10].value() : 999.0;
  out[12] = obj[11].has_value() ? obj[11].value() : 999.0;
  out[13] = obj[12].has_value() ? obj[12].value() : 999.0;
}}
"""
    )
    out = np.zeros(14, dtype=np.float64)
    ROOT.roottest_nested_BitMaskedArray_NumpyArray_v2b(
        out, len(layout), lookup.arrayptrs
    )
    assert out.tolist() == [
        13.0,
        0.0,
        1.0,
        2.0,
        3.0,
        999.0,
        999.0,
        999.0,
        999.0,
        1.1,
        999.0,
        3.3,
        999.0,
        5.5,
    ]

    v2c = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 14], dtype=np.int64)),
        ak.contents.bitmaskedarray.BitMaskedArray(
            ak.index.Index(
                np.packbits(
                    np.array(
                        [
                            0,
                            0,
                            0,
                            1,
                            1,
                            1,
                            1,
                            0,
                            0,
                            0,
                            1,
                            0,
                            1,
                            0,
                            1,
                            0,
                            0,
                        ],
                        np.uint8,
                    )
                )
            ),
            ak.contents.numpyarray.NumpyArray(
                np.array(
                    [
                        999,
                        0.0,
                        1.0,
                        2.0,
                        3.0,
                        4.0,
                        5.0,
                        6.0,
                        7.0,
                        1.1,
                        2.2,
                        3.3,
                        4.4,
                        5.5,
                        6.6,
                    ]
                )
            ),
            valid_when=True,
            length=14,
            lsb_order=True,
        ),
    )

    layout = v2c
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_nested_BitMaskedArray_NumpyArray_v2c(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()}[1];
  out[0] = obj.size();
  out[1] = obj[0].has_value() ? obj[0].value() : 999.0;
  out[2] = obj[1].has_value() ? obj[1].value() : 999.0;
  out[3] = obj[2].has_value() ? obj[2].value() : 999.0;
  out[4] = obj[3].has_value() ? obj[3].value() : 999.0;
  out[5] = obj[4].has_value() ? obj[4].value() : 999.0;
  out[6] = obj[5].has_value() ? obj[5].value() : 999.0;
  out[7] = obj[6].has_value() ? obj[6].value() : 999.0;
  out[8] = obj[7].has_value() ? obj[7].value() : 999.0;
  out[9] = obj[8].has_value() ? obj[8].value() : 999.0;
  out[10] = obj[9].has_value() ? obj[9].value() : 999.0;
  out[11] = obj[10].has_value() ? obj[10].value() : 999.0;
  out[12] = obj[11].has_value() ? obj[11].value() : 999.0;
  out[13] = obj[12].has_value() ? obj[12].value() : 999.0;
}}
"""
    )
    out = np.zeros(14, dtype=np.float64)
    ROOT.roottest_nested_BitMaskedArray_NumpyArray_v2c(
        out, len(layout), lookup.arrayptrs
    )
    assert out.tolist() == [
        13.0,
        0.0,
        1.0,
        2.0,
        3.0,
        999.0,
        999.0,
        999.0,
        999.0,
        1.1,
        999.0,
        3.3,
        999.0,
        5.5,
    ]

    v2d = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 14], dtype=np.int64)),
        ak.contents.bitmaskedarray.BitMaskedArray(
            ak.index.Index(
                np.packbits(
                    np.array(
                        [
                            1,
                            1,
                            1,
                            0,
                            0,
                            0,
                            0,
                            1,
                            1,
                            1,
                            0,
                            1,
                            0,
                            1,
                            0,
                            1,
                            1,
                        ],
                        np.uint8,
                    )
                )
            ),
            ak.contents.numpyarray.NumpyArray(
                np.array(
                    [
                        999,
                        0.0,
                        1.0,
                        2.0,
                        3.0,
                        4.0,
                        5.0,
                        6.0,
                        7.0,
                        1.1,
                        2.2,
                        3.3,
                        4.4,
                        5.5,
                        6.6,
                    ]
                )
            ),
            valid_when=False,
            length=14,
            lsb_order=True,
        ),
    )

    layout = v2d
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_nested_BitMaskedArray_NumpyArray_v2d(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()}[1];
  out[0] = obj.size();
  out[1] = obj[0].has_value() ? obj[0].value() : 999.0;
  out[2] = obj[1].has_value() ? obj[1].value() : 999.0;
  out[3] = obj[2].has_value() ? obj[2].value() : 999.0;
  out[4] = obj[3].has_value() ? obj[3].value() : 999.0;
  out[5] = obj[4].has_value() ? obj[4].value() : 999.0;
  out[6] = obj[5].has_value() ? obj[5].value() : 999.0;
  out[7] = obj[6].has_value() ? obj[6].value() : 999.0;
  out[8] = obj[7].has_value() ? obj[7].value() : 999.0;
  out[9] = obj[8].has_value() ? obj[8].value() : 999.0;
  out[10] = obj[9].has_value() ? obj[9].value() : 999.0;
  out[11] = obj[10].has_value() ? obj[10].value() : 999.0;
  out[12] = obj[11].has_value() ? obj[11].value() : 999.0;
  out[13] = obj[12].has_value() ? obj[12].value() : 999.0;
}}
"""
    )
    out = np.zeros(14, dtype=np.float64)
    ROOT.roottest_nested_BitMaskedArray_NumpyArray_v2d(
        out, len(layout), lookup.arrayptrs
    )
    assert out.tolist() == [
        13.0,
        0.0,
        1.0,
        2.0,
        3.0,
        999.0,
        999.0,
        999.0,
        999.0,
        1.1,
        999.0,
        3.3,
        999.0,
        5.5,
    ]


def test_nested_UnmaskedArray_NumpyArray():
    v2a = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 5], dtype=np.int64)),
        ak.contents.unmaskedarray.UnmaskedArray(
            ak.contents.numpyarray.NumpyArray(np.array([999, 0.0, 1.1, 2.2, 3.3]))
        ),
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_nested_UnmaskedArray_NumpyArray_v2a(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()}[1];
  out[0] = obj.size();
  out[1] = obj[1].has_value() ? obj[1].value() : 999.0;
  out[2] = obj[3].has_value() ? obj[3].value() : 999.0;
}}
"""
    )
    out = np.zeros(3, dtype=np.float64)
    ROOT.roottest_nested_UnmaskedArray_NumpyArray_v2a(
        out, len(layout), lookup.arrayptrs
    )
    assert out.tolist() == [4.0, 1.1, 3.3]


def test_nested_UnionArray_NumpyArray():
    v2a = ak.contents.ListOffsetArray(
        ak.index.Index64(np.array([0, 1, 8], dtype=np.int64)),
        ak.contents.unionarray.UnionArray(
            ak.index.Index(np.array([123, 1, 1, 0, 0, 1, 0, 1], np.int8)),
            ak.index.Index(np.array([999, 4, 3, 0, 1, 2, 2, 4, 100], np.int64)),
            [
                ak.from_iter(["1", "2", "3"], highlevel=False),
                ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5])),
            ],
        ),
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_nested_UnionArray_NumpyArray_v2a(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()}[1];
  out[0] = obj.size();
  out[1] = obj[0].index();
  out[2] = obj[1].index();
  out[3] = obj[2].index();
  out[4] = obj[3].index();
  out[5] = obj[4].index();
  out[6] = obj[5].index();
  out[7] = obj[6].index();
  out[8] = std::get<double>(obj[0]);
  out[9] = std::get<double>(obj[1]);
  out[10] = std::atoi(std::get<std::string>(obj[2]).c_str());
  out[11] = std::atoi(std::get<std::string>(obj[3]).c_str());
  out[12] = std::get<double>(obj[4]);
  out[13] = std::atoi(std::get<std::string>(obj[5]).c_str());
  out[14] = std::get<double>(obj[6]);
}}
"""
    )
    out = np.zeros(15, dtype=np.float64)
    ROOT.roottest_nested_UnionArray_NumpyArray_v2a(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [7, 1, 1, 0, 0, 1, 0, 1, 5.5, 4.4, 1, 2, 3.3, 3, 5.5]


def test_ListArray_strings():
    layout = ak.operations.from_iter(
        ["one", "two", "three", "four", "five"], highlevel=False
    )

    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_ListArray_strings(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};
  out[0] = obj.size();
  out[1] = (obj[0] == "one");
  out[2] = (obj[1] == "two");
  out[3] = (obj[2] == "three");
  out[4] = (obj[3] == "four");
  out[5] = (obj[4] == "five");
}}
"""
    )
    out = np.zeros(6, dtype=np.float64)
    ROOT.roottest_ListArray_strings(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [5.0, 1.0, 1.0, 1.0, 1.0, 1.0]


def test_RegularArray_strings():
    layout = ak.operations.to_regular(
        ak.operations.from_iter(["onexx", "twoxx", "three", "fourx", "fivex"]),
        axis=1,
        highlevel=False,
    )

    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_RegularArray_strings(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};
  out[0] = obj.size();
  out[1] = (obj[0] == "onexx");
  out[2] = (obj[1] == "twoxx");
  out[3] = (obj[2] == "three");
  out[4] = (obj[3] == "fourx");
  out[5] = (obj[4] == "fivex");
}}
"""
    )
    out = np.zeros(6, dtype=np.float64)
    ROOT.roottest_RegularArray_strings(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [5.0, 1.0, 1.0, 1.0, 1.0, 1.0]


def test_NumpyArray_iterator():
    v2a = ak.contents.numpyarray.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3]),
        parameters={"some": "stuff", "other": [1, 2, "three"]},
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_NumpyArray_iterator_v2a(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};

  size_t i = 0;
  for (auto it = obj.begin();  it != obj.end();  ++it) {{
    out[i] = *it;
    i++;
  }}
}}
"""
    )
    out = np.zeros(4, dtype=np.float64)
    ROOT.roottest_NumpyArray_iterator_v2a(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [0.0, 1.1, 2.2, 3.3]


def test_NumpyArray_iterator2():
    v2a = ak.contents.numpyarray.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3]),
        parameters={"some": "stuff", "other": [1, 2, "three"]},
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_NumpyArray_iterator2_v2a(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};

  size_t i = 0;
  for (auto x : obj) {{
    out[i] = x;
    i++;
  }}
}}
"""
    )
    out = np.zeros(4, dtype=np.float64)
    ROOT.roottest_NumpyArray_iterator2_v2a(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [0.0, 1.1, 2.2, 3.3]


def test_NumpyArray_riterator():
    v2a = ak.contents.numpyarray.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3]),
        parameters={"some": "stuff", "other": [1, 2, "three"]},
    )

    layout = v2a
    generator = ak._connect.cling.togenerator(layout.form, flatlist_as_rvec=False)
    lookup = ak._lookup.Lookup(layout, generator)
    generator.generate(compiler)

    ROOT.gInterpreter.Declare(
        f"""
void roottest_NumpyArray_riterator_v2a(double* out, ssize_t length, ssize_t* ptrs) {{
  auto obj = {generator.dataset()};

  size_t i = 0;
  for (auto it = obj.rbegin();  it != obj.rend();  ++it) {{
    out[i] = *it;
    i++;
  }}
}}
"""
    )
    out = np.zeros(4, dtype=np.float64)
    ROOT.roottest_NumpyArray_riterator_v2a(out, len(layout), lookup.arrayptrs)
    assert out.tolist() == [3.3, 2.2, 1.1, 0.0]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE


from numba import types

import awkward as ak
from awkward._backends.cupy import CupyBackend

########## ArrayView Arguments Handler for CUDA JIT


class ArrayViewArgHandler:
    def prepare_args(self, ty, val, stream, retr):
        if isinstance(val, ak.Array):
            if isinstance(val.layout.backend, CupyBackend):
                # Use uint64 for pos, start, stop, the array pointers values, and the pylookup value
                tys = types.UniTuple(types.uint64, 5)

                start = val._numbaview.start
                stop = val._numbaview.stop
                pos = val._numbaview.pos
                arrayptrs = val._numbaview.lookup.arrayptrs.data.ptr
                pylookup = 0

                return tys, (pos, start, stop, arrayptrs, pylookup)
            else:
                raise TypeError(
                    '`ak.to_backend` should be called with `backend="cuda"` to put '
                    "the array on the GPU before using it: "
                    'ak.to_backend(array, backend="cuda")'
                )

        else:
            return ty, val



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test():
    x = ak.contents.ByteMaskedArray(
        ak.index.Index8(np.r_[0, 1, 1]),
        ak.contents.NumpyArray(np.arange(12)),
        valid_when=True,
    )
    y = ak.contents.ByteMaskedArray(
        ak.index.Index8(np.r_[1, 1, 1, 0, 0]),
        ak.contents.NumpyArray(np.arange(12)),
        valid_when=True,
    )
    z = x._mergemany([y])

    assert z.to_list() == [None, 1, 2, 0, 1, 2, None, None]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np  # noqa: F401
import pytest

import awkward as ak

pytest.importorskip("pyarrow")


def test():
    layout = ak.contents.RecordArray([], fields=[], length=3)
    struct = ak.to_arrow(layout)
    assert len(struct) == len(layout) == 3
    assert struct.tolist() == [{}, {}, {}]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE

import numpy as np
import pytest  # noqa: F401

import awkward as ak


def test_byte_masked_array():
    array = ak.highlevel.Array(
        ak.contents.ByteMaskedArray(
            ak.index.Index8(np.array([True, True])),
            ak.contents.RecordArray(
                [
                    ak.contents.NumpyArray(np.arange(2)),
                    ak.contents.IndexedOptionArray(
                        ak.index.Index64(np.array([0, -1])),
                        ak.contents.NumpyArray(np.array([1], dtype=np.int64)),
                    ),
                ],
                ["a", "b"],
            ),
            # ak.Array({"a": [1, 2], "b": [1, None]}).layout,
            valid_when=True,
        )
    )

    assert str(array.b.type) == "2 * ?int64"


def test_bit_masked_array():
    array = ak.highlevel.Array(
        ak.contents.BitMaskedArray(
            ak.index.IndexU8(np.array([1 << 0 | 1 << 1], dtype=np.uint8)),
            ak.contents.RecordArray(
                [
                    ak.contents.NumpyArray(np.arange(2)),
                    ak.contents.IndexedOptionArray(
                        ak.index.Index64(np.array([0, -1])),
                        ak.contents.NumpyArray(np.array([1], dtype=np.int64)),
                    ),
                ],
                ["a", "b"],
            ),
            valid_when=True,
            length=2,
            lsb_order=True,
        )
    )

    assert str(array.b.type) == "2 * ?int64"



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
# pylint: disable=consider-using-enumerate
from __future__ import annotations

import copy
import ctypes
from collections.abc import Iterable, MutableMapping, Sequence

import awkward as ak
from awkward._backends.backend import Backend
from awkward._errors import AxisError, deprecate
from awkward._layout import maybe_posaxis
from awkward._nplikes.numpy import Numpy
from awkward._nplikes.numpylike import ArrayLike, IndexType, NumpyMetadata
from awkward._nplikes.shape import ShapeItem, unknown_length
from awkward._nplikes.typetracer import OneOf, TypeTracer
from awkward._parameters import parameters_intersect, parameters_union
from awkward._regularize import is_integer_like
from awkward._slicing import NO_HEAD
from awkward._typing import TYPE_CHECKING, Callable, Final, Self, SupportsIndex, final
from awkward._util import UNSET
from awkward.contents.content import Content
from awkward.forms.form import Form
from awkward.forms.unionform import UnionForm
from awkward.index import Index, Index8, Index64

if TYPE_CHECKING:
    from awkward._slicing import SliceItem

np = NumpyMetadata.instance()
numpy = Numpy.instance()


@final
class UnionArray(Content):
    """
    UnionArray represents data drawn from an ordered list of `contents`,
    which can have different types, using

    * `tags`: buffer of integers indicating which content each array element draws from.
    * `index`: buffer of integers indicating which element from the content to draw from.

    UnionArrays correspond to Apache Arrow's
    [dense union type](https://arrow.apache.org/docs/format/Columnar.html#dense-union).
    Awkward Array has no direct equivalent for Apache Arrow's
    [sparse union type](https://arrow.apache.org/docs/format/Columnar.html#sparse-union).

    To illustrate how the constructor arguments are interpreted, the following is a
    simplified implementation of `__init__`, `__len__`, and `__getitem__`:

        class UnionArray(Content):
            def __init__(self, tags, index, contents):
                assert isinstance(tags, Index8)
                assert isinstance(index, (Index32, IndexU32, Index64))
                assert isinstance(contents, list)
                assert len(index) >= len(tags)  # usually equal
                for x in tags:
                    assert 0 <= x < len(contents)
                for i, x in enumerate(tags):
                    assert 0 <= index[i] < len(contents[x])
                self.tags = tags
                self.index = index
                self.contents = contents

            def __len__(self):
                return len(self.tags)

            def __getitem__(self, where):
                if isinstance(where, int):
                    if where < 0:
                        where += len(self)
                    assert 0 <= where < len(self)
                    return self.contents[self.tags[where]][self.index[where]]

                elif isinstance(where, slice) and where.step is None:
                    return UnionArray(
                        self.tags[where], self.index[where], self.contents
                    )

                elif isinstance(where, str):
                    return UnionArray(
                        self.tags, self.index, [x[where] for x in self.contents]
                    )

                else:
                    raise AssertionError(where)
    """

    is_union = True

    def __init__(self, tags, index, contents, *, parameters=None):
        if not (isinstance(tags, Index) and tags.dtype == np.dtype(np.int8)):
            raise TypeError(
                "{} 'tags' must be an Index with dtype=int8, not {}".format(
                    type(self).__name__, repr(tags)
                )
            )

        if not isinstance(index, Index) and index.dtype in (
            np.dtype(np.int32),
            np.dtype(np.uint32),
            np.dtype(np.int64),
        ):
            raise TypeError(
                "{} 'index' must be an Index with dtype in (int32, uint32, int64), "
                "not {}".format(type(self).__name__, repr(index))
            )

        if not isinstance(contents, Iterable):
            raise TypeError(
                "{} 'contents' must be iterable, not {}".format(
                    type(self).__name__, repr(contents)
                )
            )
        if not isinstance(contents, list):
            contents = list(contents)

        if len(contents) < 2:
            raise TypeError(f"{type(self).__name__} must have at least 2 'contents'")

        n_seen_options = 0
        for content in contents:
            if not isinstance(content, Content):
                raise TypeError(
                    "{} all 'contents' must be Content subclasses, not {}".format(
                        type(self).__name__, repr(content)
                    )
                )
            if content.is_union:
                raise TypeError(
                    "{0} cannot contain union-types in its 'contents' ({1}); try {0}.simplified instead".format(
                        type(self).__name__, type(content).__name__
                    )
                )
            elif content.is_option:
                n_seen_options += 1
            elif content.is_indexed and content.parameter("__array__") != "categorical":
                raise TypeError(
                    (
                        "{0} cannot contain non-categorical indexed-types in its 'contents' ({1}); "
                        "try {0}.simplified instead"
                    ).format(type(self).__name__, type(content).__name__)
                )

        if n_seen_options not in {0, len(contents)}:
            raise TypeError(
                "{0} must either be comprised of entirely optional contents, or no optional contents; "
                "try {0}.simplified instead".format(type(self).__name__)
            )

        if (
            not (tags.length is unknown_length or index.length is unknown_length)
            and tags.length > index.length
        ):
            raise ValueError(
                "{} len(tags) ({}) must be <= len(index) ({})".format(
                    type(self).__name__, tags.length, index.length
                )
            )

        backend = None
        for content in contents:
            if backend is None:
                backend = content.backend
            elif backend is not content.backend:
                raise TypeError(
                    "{} 'contents' must use the same array library (backend): {} vs {}".format(
                        type(self).__name__,
                        type(backend).__name__,
                        type(content.backend).__name__,
                    )
                )

        assert tags.nplike is backend.index_nplike
        assert index.nplike is backend.index_nplike

        self._tags = tags
        self._index = index
        self._contents = contents
        self._init(parameters, backend)

    @property
    def tags(self):
        return self._tags

    @property
    def index(self):
        return self._index

    @property
    def contents(self):
        return self._contents

    form_cls: Final = UnionForm

    def copy(
        self,
        tags=UNSET,
        index=UNSET,
        contents=UNSET,
        *,
        parameters=UNSET,
    ):
        return UnionArray(
            self._tags if tags is UNSET else tags,
            self._index if index is UNSET else index,
            self._contents if contents is UNSET else contents,
            parameters=self._parameters if parameters is UNSET else parameters,
        )

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        return self.copy(
            tags=copy.deepcopy(self._tags, memo),
            index=copy.deepcopy(self._index, memo),
            contents=[copy.deepcopy(x, memo) for x in self._contents],
            parameters=copy.deepcopy(self._parameters, memo),
        )

    @classmethod
    def simplified(
        cls,
        tags,
        index,
        contents,
        *,
        parameters=None,
        merge=UNSET,
        mergebool=False,
    ):
        if merge is UNSET:
            merge = True
        else:
            deprecate(
                "The `merge` argument to UnionArray.simplified(...) is deprecated; "
                "it is no longer possible to construct UnionArrays with mergeable contents.",
                version="2.4.0",
            )
        self_index = index
        self_tags = tags
        self_contents = contents

        backend = None
        for content in contents:
            if backend is None:
                backend = content.backend
            elif backend is not content.backend:
                raise TypeError(
                    "{} 'contents' must use the same array library (backend): {} vs {}".format(
                        cls.__name__,
                        type(backend).__name__,
                        type(content.backend).__name__,
                    )
                )

        if backend.nplike.known_data and self_index.length < self_tags.length:
            raise ValueError("invalid UnionArray: len(index) < len(tags)")

        length = self_tags.length
        tags = ak.index.Index8.empty(length, backend.index_nplike)
        index = ak.index.Index64.empty(length, backend.index_nplike)
        contents = []

        # For each outer union content
        for i, self_cont in enumerate(self_contents):
            # Is one of our new contents also a union?
            if isinstance(self_cont, UnionArray):
                innertags = self_cont._tags
                innerindex = self_cont._index
                innercontents = self_cont._contents

                # Update outermost parameters with this union's parameters
                parameters = parameters_union(self_cont._parameters, parameters)

                # For each inner union content
                for j, inner_cont in enumerate(innercontents):
                    unmerged = True

                    # For each "final" outer union content
                    for k in range(len(contents)):
                        # Try and merge inner union content with running outer-union contentca
                        if merge and contents[k]._mergeable_next(inner_cont, mergebool):
                            backend.maybe_kernel_error(
                                backend[
                                    "awkward_UnionArray_simplify",
                                    tags.dtype.type,
                                    index.dtype.type,
                                    self_tags.dtype.type,
                                    self_index.dtype.type,
                                    innertags.dtype.type,
                                    innerindex.dtype.type,
                                ](
                                    tags.data,
                                    index.data,
                                    self_tags.data,
                                    self_index.data,
                                    innertags.data,
                                    innerindex.data,
                                    k,
                                    j,
                                    i,
                                    length,
                                    contents[k].length,
                                )
                            )
                            contents[k] = contents[k]._mergemany([inner_cont])
                            unmerged = False
                            break

                    # Did we fail to merge any of the final outer contents with this inner union content?
                    if unmerged:
                        backend.maybe_kernel_error(
                            backend[
                                "awkward_UnionArray_simplify",
                                tags.dtype.type,
                                index.dtype.type,
                                self_tags.dtype.type,
                                self_index.dtype.type,
                                innertags.dtype.type,
                                innerindex.dtype.type,
                            ](
                                tags.data,
                                index.data,
                                self_tags.data,
                                self_index.data,
                                innertags.data,
                                innerindex.data,
                                len(contents),
                                j,
                                i,
                                length,
                                0,
                            )
                        )
                        contents.append(inner_cont)

            else:
                unmerged = True
                for k in range(len(contents)):
                    if contents[k] is self_cont:
                        backend.maybe_kernel_error(
                            backend[
                                "awkward_UnionArray_simplify_one",
                                tags.dtype.type,
                                index.dtype.type,
                                self_tags.dtype.type,
                                self_index.dtype.type,
                            ](
                                tags.data,
                                index.data,
                                self_tags.data,
                                self_index.data,
                                k,
                                i,
                                length,
                                0,
                            )
                        )
                        unmerged = False
                        break

                    elif merge and contents[k]._mergeable_next(self_cont, mergebool):
                        backend.maybe_kernel_error(
                            backend[
                                "awkward_UnionArray_simplify_one",
                                tags.dtype.type,
                                index.dtype.type,
                                self_tags.dtype.type,
                                self_index.dtype.type,
                            ](
                                tags.data,
                                index.data,
                                self_tags.data,
                                self_index.data,
                                k,
                                i,
                                length,
                                contents[k].length,
                            )
                        )
                        contents[k] = contents[k]._mergemany([self_cont])
                        unmerged = False
                        break

                if unmerged:
                    backend.maybe_kernel_error(
                        backend[
                            "awkward_UnionArray_simplify_one",
                            tags.dtype.type,
                            index.dtype.type,
                            self_tags.dtype.type,
                            self_index.dtype.type,
                        ](
                            tags.data,
                            index.data,
                            self_tags.data,
                            self_index.data,
                            len(contents),
                            i,
                            length,
                            0,
                        )
                    )
                    contents.append(self_cont)

        if len(contents) > 2**7:
            raise NotImplementedError(
                "FIXME: handle UnionArray with more than 127 contents"
            )

        # If any contents are non-categorical index types, we can merge them into the union
        # This is safe, because any remaining index types at this point in the routine are not considered
        # mergeable with the other contents. This means none of the other contents are option or index types,
        # nor are any contents mergeable with these index types' contents. We already know that the
        # other contents cannot be unions. Thus, it's safe to simply erase the outer indexed type.
        for tag, content in enumerate(contents):
            if (
                isinstance(content, ak.contents.IndexedArray)
                and content.parameter("__array__") != "categorical"
            ):
                # only want to affect the part of the UnionArray.index for this tag
                selection = tags.data == tag
                # function-composition of this part of the UnionArray.index with the IndexedArray.index
                index.data[selection] = content.index.data[index.data[selection]]
                # now we don't have an IndexedArray anymore, but we want to preserve its parameters
                contents[tag] = content.content.copy(
                    parameters=parameters_union(
                        content.content.parameters, content.parameters
                    )
                )

        # If any contents are options, ensure all contents are options!
        if any(c.is_option for c in contents):
            contents = [
                c
                if c.is_option
                else c.to_IndexedOptionArray64()
                if c.is_indexed
                else ak.contents.UnmaskedArray(c)
                for c in contents
            ]

        if len(contents) == 1:
            next = contents[0]._carry(index, True)
            return next.copy(parameters=parameters_union(next._parameters, parameters))

        else:
            return cls(
                tags,
                index,
                contents,
                parameters=parameters,
            )

    def content(self, index):
        return self._contents[index]

    def _form_with_key(self, getkey: Callable[[Content], str | None]) -> UnionForm:
        form_key = getkey(self)
        return self.form_cls(
            self._tags.form,
            self._index.form,
            [x._form_with_key(getkey) for x in self._contents],
            parameters=self._parameters,
            form_key=form_key,
        )

    def _to_buffers(
        self,
        form: Form,
        getkey: Callable[[Content, Form, str], str],
        container: MutableMapping[str, ArrayLike],
        backend: Backend,
        byteorder: str,
    ):
        assert isinstance(form, self.form_cls)
        key1 = getkey(self, form, "tags")
        key2 = getkey(self, form, "index")
        container[key1] = ak._util.native_to_byteorder(
            self._tags.raw(backend.index_nplike), byteorder
        )
        container[key2] = ak._util.native_to_byteorder(
            self._index.raw(backend.index_nplike), byteorder
        )
        for i, content in enumerate(self._contents):
            content._to_buffers(form.content(i), getkey, container, backend, byteorder)

    def _to_typetracer(self, forget_length: bool) -> Self:
        tt = TypeTracer.instance()
        tags = self._tags.to_nplike(tt)
        return UnionArray(
            tags.forget_length() if forget_length else tags,
            self._index.to_nplike(tt),
            [x._to_typetracer(False) for x in self._contents],
            parameters=self._parameters,
        )

    def _touch_data(self, recursive: bool):
        self._tags._touch_data()
        self._index._touch_data()
        if recursive:
            for x in self._contents:
                x._touch_data(recursive)

    def _touch_shape(self, recursive: bool):
        self._tags._touch_shape()
        self._index._touch_shape()
        if recursive:
            for x in self._contents:
                x._touch_shape(recursive)

    @property
    def length(self) -> ShapeItem:
        return self._tags.length

    def __repr__(self):
        return self._repr("", "", "")

    def _repr(self, indent, pre, post):
        out = [indent, pre, "<UnionArray len="]
        out.append(repr(str(self.length)))
        out.append(">")
        out.extend(self._repr_extra(indent + "    "))
        out.append("\n")
        out.append(self._tags._repr(indent + "    ", "<tags>", "</tags>\n"))
        out.append(self._index._repr(indent + "    ", "<index>", "</index>\n"))

        for i, x in enumerate(self._contents):
            out.append(f"{indent}    <content index={str(i)!r}>\n")
            out.append(x._repr(indent + "        ", "", "\n"))
            out.append(f"{indent}    </content>\n")

        out.append(indent + "</UnionArray>")
        out.append(post)
        return "".join(out)

    def _getitem_nothing(self):
        return self._getitem_range(0, 0)

    def _getitem_at(self, where: IndexType):
        if not self._backend.nplike.known_data:
            self._touch_data(recursive=False)
            return OneOf([x._getitem_at(where) for x in self._contents])

        if where < 0:
            where += self.length
        if self._backend.nplike.known_data and not 0 <= where < self.length:
            raise ak._errors.index_error(self, where)
        tag, index = self._tags[where], self._index[where]
        return self._contents[tag]._getitem_at(index)

    def _getitem_range(self, start: SupportsIndex, stop: IndexType) -> Content:
        if not self._backend.nplike.known_data:
            self._touch_shape(recursive=False)
            return self

        return UnionArray(
            self._tags[start:stop],
            self._index[start:stop],
            self._contents,
            parameters=self._parameters,
        )

    def _getitem_field(
        self, where: str | SupportsIndex, only_fields: tuple[str, ...] = ()
    ) -> Content:
        return UnionArray.simplified(
            self._tags,
            self._index,
            [x._getitem_field(where, only_fields) for x in self._contents],
            parameters=None,
        )

    def _getitem_fields(
        self, where: list[str | SupportsIndex], only_fields: tuple[str, ...] = ()
    ) -> Content:
        return UnionArray.simplified(
            self._tags,
            self._index,
            [x._getitem_fields(where, only_fields) for x in self._contents],
            parameters=None,
        )

    def _carry(self, carry: Index, allow_lazy: bool) -> Content:
        assert isinstance(carry, ak.index.Index)

        try:
            nexttags = self._tags[carry.data]
            nextindex = self._index[: self._tags.length][carry.data]
        except IndexError as err:
            raise ak._errors.index_error(self, carry.data, str(err)) from err

        return UnionArray(
            nexttags,
            nextindex,
            self._contents,
            parameters=self._parameters,
        )

    def _union_of_optionarrays(self, index, parameters):
        tag_for_missing = 0
        for i, content in enumerate(self._contents):
            if content.is_option:
                tag_for_missing = i
                break

        if not self._backend.nplike.known_data:
            self._touch_data(recursive=False)
            nexttags = self._tags.data
            nextindex = self._index.data
            contents = []
            for tag, content in enumerate(self._contents):
                if tag == tag_for_missing:
                    indexedoption_index = self._backend.index_nplike.arange(
                        content.length + 1, dtype=np.int64
                    )
                    contents.append(
                        ak.contents.IndexedOptionArray.simplified(
                            ak.index.Index64(indexedoption_index), content
                        )
                    )
                else:
                    contents.append(ak.contents.UnmaskedArray.simplified(content))

        else:
            # like _carry, above
            carry_data = index.raw(self._backend.nplike).copy()
            is_missing = carry_data < 0

            if self._tags.length != 0:
                # but the missing values will temporarily use 0 as a placeholder
                carry_data[is_missing] = 0
                try:
                    nexttags = self._tags.data[carry_data]
                    nextindex = self._index.data[: self._tags.length][carry_data]
                except IndexError as err:
                    raise ak._errors.index_error(self, carry_data, str(err)) from err

                # now actually set the missing values
                nexttags[is_missing] = tag_for_missing
                nextindex[is_missing] = self._contents[tag_for_missing].length

            else:
                # UnionArray is empty, so
                nexttags = self._backend.index_nplike.full(
                    len(carry_data), tag_for_missing, dtype=self._tags.dtype
                )
                nextindex = self._backend.index_nplike.full(
                    len(carry_data),
                    self._contents[tag_for_missing].length,
                    dtype=self._index.dtype,
                )

            contents = []
            for tag, content in enumerate(self._contents):
                if tag == tag_for_missing:
                    indexedoption_index = self._backend.index_nplike.arange(
                        content.length + 1, dtype=np.int64
                    )
                    indexedoption_index[content.length] = -1
                    contents.append(
                        ak.contents.IndexedOptionArray.simplified(
                            ak.index.Index64(indexedoption_index), content
                        )
                    )
                else:
                    contents.append(ak.contents.UnmaskedArray.simplified(content))

        return UnionArray.simplified(
            ak.index.Index(nexttags),
            ak.index.Index(nextindex),
            contents,
            parameters=parameters_union(self._parameters, parameters),
        )

    def project(self, index):
        lentags = self._tags.length
        assert (
            self._index.length is unknown_length or lentags is unknown_length
        ) or self._index.length >= lentags
        lenout = ak.index.Index64.empty(1, self._backend.index_nplike)
        tmpcarry = ak.index.Index64.empty(lentags, self._backend.index_nplike)
        assert (
            lenout.nplike is self._backend.index_nplike
            and tmpcarry.nplike is self._backend.index_nplike
            and self._tags.nplike is self._backend.index_nplike
            and self._index.nplike is self._backend.index_nplike
        )
        self._backend.maybe_kernel_error(
            self._backend[
                "awkward_UnionArray_project",
                lenout.dtype.type,
                tmpcarry.dtype.type,
                self._tags.dtype.type,
                self._index.dtype.type,
            ](
                lenout.data,
                tmpcarry.data,
                self._tags.data,
                self._index.data,
                lentags,
                index,
            )
        )
        nextcarry = ak.index.Index64(
            tmpcarry.data[: lenout[0]], nplike=self._backend.index_nplike
        )
        return self._contents[index]._carry(nextcarry, False)

    @staticmethod
    def regular_index(
        tags: Index,
        *,
        backend: Backend,
        index_cls: type[Index] = Index64,
    ):
        tags = tags.to_nplike(backend.index_nplike)

        lentags = tags.length
        _size = ak.index.Index64.empty(1, nplike=backend.index_nplike)
        assert (
            _size.nplike is backend.index_nplike and tags.nplike is backend.index_nplike
        )
        backend.maybe_kernel_error(
            backend[
                "awkward_UnionArray_regular_index_getsize",
                _size.dtype.type,
                tags.dtype.type,
            ](
                _size.data,
                tags.data,
                lentags,
            )
        )
        size = backend.index_nplike.index_as_shape_item(_size[0])
        current = index_cls.empty(size, nplike=backend.index_nplike)
        outindex = index_cls.empty(lentags, nplike=backend.index_nplike)
        assert (
            outindex.nplike is backend.index_nplike
            and current.nplike is backend.index_nplike
            and tags.nplike is backend.index_nplike
        )
        backend.maybe_kernel_error(
            backend[
                "awkward_UnionArray_regular_index",
                outindex.dtype.type,
                current.dtype.type,
                tags.dtype.type,
            ](
                outindex.data,
                current.data,
                size,
                tags.data,
                lentags,
            )
        )
        return outindex

    def _regular_index(self, tags: Index) -> Index:
        return self.regular_index(
            tags, index_cls=type(self._index), backend=self._backend
        )

    @staticmethod
    def nested_tags_index(
        offsets: Index,
        counts: Sequence[Index],
        *,
        backend: Backend,
        tags_cls: type[Index] = Index8,
        index_cls: type[Index] = Index64,
    ) -> tuple[Index, Index]:
        index_nplike = backend.index_nplike

        offsets = offsets.to_nplike(index_nplike)
        counts = [c.to_nplike(index_nplike) for c in counts]

        f_offsets = ak.index.Index64(index_nplike.asarray(offsets.data, copy=True))
        contentlen = index_nplike.index_as_shape_item(
            f_offsets[index_nplike.shape_item_as_index(f_offsets.length - 1)]
        )

        tags = tags_cls.empty(contentlen, nplike=index_nplike)
        index = index_cls.empty(contentlen, nplike=index_nplike)

        for tag, count in enumerate(counts):
            assert (
                tags.nplike is index_nplike
                and index.nplike is index_nplike
                and f_offsets.nplike is index_nplike
                and count.nplike is index_nplike
            )
            backend.maybe_kernel_error(
                backend[
                    "awkward_UnionArray_nestedfill_tags_index",
                    tags.dtype.type,
                    index.dtype.type,
                    f_offsets.dtype.type,
                    count.dtype.type,
                ](
                    tags.data,
                    index.data,
                    f_offsets.data,
                    tag,
                    count.data,
                    f_offsets.length - 1,
                )
            )
        return (tags, index)

    def _getitem_next_jagged_generic(self, slicestarts, slicestops, slicecontent, tail):
        if isinstance(self, ak.contents.UnionArray):
            raise ak._errors.index_error(
                self,
                ak.contents.ListArray(
                    slicestarts, slicestops, slicecontent, parameters=None
                ),
                "cannot apply jagged slices to irreducible union arrays",
            )
        return self._getitem_next_jagged(slicestarts, slicestops, slicecontent, tail)

    def _getitem_next_jagged(
        self, slicestarts: Index, slicestops: Index, slicecontent: Content, tail
    ) -> Content:
        return self._getitem_next_jagged_generic(
            slicestarts, slicestops, slicecontent, tail
        )

    def _getitem_next(
        self,
        head: SliceItem | tuple,
        tail: tuple[SliceItem, ...],
        advanced: Index | None,
    ) -> Content:
        if head is NO_HEAD:
            return self

        elif is_integer_like(head) or isinstance(
            head, (slice, ak.index.Index64, ak.contents.ListOffsetArray)
        ):
            outcontents = []
            for i in range(len(self._contents)):
                projection = self.project(i)
                outcontents.append(projection._getitem_next(head, tail, advanced))
            outindex = self._regular_index(self._tags)

            return UnionArray.simplified(
                self._tags,
                outindex,
                outcontents,
                parameters=self._parameters,
            )

        elif isinstance(head, str):
            return self._getitem_next_field(head, tail, advanced)

        elif isinstance(head, list):
            return self._getitem_next_fields(head, tail, advanced)

        elif head is np.newaxis:
            return self._getitem_next_newaxis(tail, advanced)

        elif head is Ellipsis:
            return self._getitem_next_ellipsis(tail, advanced)

        elif isinstance(head, ak.contents.IndexedOptionArray):
            return self._getitem_next_missing(head, tail, advanced)

        else:
            raise AssertionError(repr(head))

    def _offsets_and_flattened(self, axis: int, depth: int) -> tuple[Index, Content]:
        posaxis = maybe_posaxis(self, axis, depth)

        if posaxis is not None and posaxis + 1 == depth:
            raise AxisError("axis=0 not allowed for flatten")

        else:
            has_offsets = False
            offsetsraws = self._backend.index_nplike.empty(
                len(self._contents), dtype=np.intp
            )
            contents = []

            for i in range(len(self._contents)):
                offsets, flattened = self._contents[i]._offsets_and_flattened(
                    axis, depth
                )
                offsetsraws[i] = offsets.ptr
                contents.append(flattened)
                has_offsets = offsets.length != 0

            if has_offsets:
                total_length = ak.index.Index64.empty(
                    1, nplike=self._backend.index_nplike
                )
                assert (
                    total_length.nplike is self._backend.index_nplike
                    and self._tags.nplike is self._backend.index_nplike
                    and self._index.nplike is self._backend.index_nplike
                )
                self._backend.maybe_kernel_error(
                    self._backend[
                        "awkward_UnionArray_flatten_length",
                        total_length.dtype.type,
                        self._tags.dtype.type,
                        self._index.dtype.type,
                        np.int64,
                    ](
                        total_length.data,
                        self._tags.data,
                        self._index.data,
                        self._tags.length,
                        offsetsraws.ctypes.data_as(
                            ctypes.POINTER(ctypes.POINTER(ctypes.c_int64))
                        ),
                    )
                )

                totags = ak.index.Index8.empty(
                    total_length[0], nplike=self._backend.index_nplike
                )
                toindex = ak.index.Index64.empty(
                    total_length[0], nplike=self._backend.index_nplike
                )
                tooffsets = ak.index.Index64.empty(
                    self._tags.length + 1, nplike=self._backend.index_nplike
                )

                assert (
                    totags.nplike is self._backend.index_nplike
                    and toindex.nplike is self._backend.index_nplike
                    and tooffsets.nplike is self._backend.index_nplike
                    and self._tags.nplike is self._backend.index_nplike
                    and self._index.nplike is self._backend.index_nplike
                )
                self._backend.maybe_kernel_error(
                    self._backend[
                        "awkward_UnionArray_flatten_combine",
                        totags.dtype.type,
                        toindex.dtype.type,
                        tooffsets.dtype.type,
                        self._tags.dtype.type,
                        self._index.dtype.type,
                        np.int64,
                    ](
                        totags.data,
                        toindex.data,
                        tooffsets.data,
                        self._tags.data,
                        self._index.data,
                        self._tags.length,
                        offsetsraws.ctypes.data_as(
                            ctypes.POINTER(ctypes.POINTER(ctypes.c_int64))
                        ),
                    )
                )

                return (
                    tooffsets,
                    UnionArray(
                        totags,
                        toindex,
                        contents,
                        parameters=self._parameters,
                    ),
                )

            else:
                offsets = ak.index.Index64.zeros(
                    0,
                    nplike=self._backend.index_nplike,
                    dtype=np.int64,
                )
                return (
                    offsets,
                    UnionArray(
                        self._tags,
                        self._index,
                        contents,
                        parameters=self._parameters,
                    ),
                )

    def _mergeable_next(self, other: Content, mergebool: bool) -> bool:
        return True

    def _merging_strategy(self, others):
        if len(others) == 0:
            raise ValueError(
                "to merge this array with 'others', at least one other must be provided"
            )

        head = [self]
        tail = []

        for i in range(len(others)):
            head.append(others[i])

        if any(isinstance(x.backend.nplike, TypeTracer) for x in head + tail):
            head = [
                x if isinstance(x.backend.nplike, TypeTracer) else x.to_typetracer()
                for x in head
            ]
            tail = [
                x if isinstance(x.backend.nplike, TypeTracer) else x.to_typetracer()
                for x in tail
            ]

        return (head, tail)

    def _reverse_merge(self, other):
        theirlength = other.length
        mylength = self.length

        tags = ak.index.Index8.empty(
            theirlength + mylength,
            nplike=self._backend.index_nplike,
        )
        index = ak.index.Index64.empty(
            theirlength + mylength,
            nplike=self._backend.index_nplike,
        )

        contents = [other]
        contents.extend(self.contents)

        assert tags.nplike is self._backend.index_nplike
        self._backend.maybe_kernel_error(
            self._backend["awkward_UnionArray_filltags_const", tags.dtype.type](
                tags.data,
                0,
                theirlength,
                0,
            )
        )

        assert index.nplike is self._backend.index_nplike
        self._backend.maybe_kernel_error(
            self._backend["awkward_UnionArray_fillindex_count", index.dtype.type](
                index.data,
                0,
                theirlength,
            )
        )

        assert (
            tags.nplike is self._backend.index_nplike
            and self.tags.nplike is self._backend.index_nplike
        )
        self._backend.maybe_kernel_error(
            self._backend[
                "awkward_UnionArray_filltags",
                tags.dtype.type,
                self.tags.dtype.type,
            ](
                tags.data,
                theirlength,
                self.tags.data,
                mylength,
                1,
            )
        )

        assert (
            index.nplike is self._backend.index_nplike
            and self.index.nplike is self._backend.index_nplike
        )
        self._backend.maybe_kernel_error(
            self._backend[
                "awkward_UnionArray_fillindex",
                index.dtype.type,
                self.index.dtype.type,
            ](
                index.data,
                theirlength,
                self.index.data,
                mylength,
            )
        )

        if len(contents) > 2**7:
            raise AssertionError("FIXME: handle UnionArray with more than 127 contents")

        return ak.contents.UnionArray.simplified(
            tags, index, contents, parameters=self._parameters
        )

    def _mergemany(self, others: Sequence[Content]) -> Content:
        if len(others) == 0:
            return self

        index_nplike = self._backend.index_nplike
        head, tail = self._merging_strategy(others)

        total_length = 0
        for array in head:
            total_length += array.length

        nexttags = ak.index.Index8.empty(total_length, nplike=index_nplike)
        nextindex = ak.index.Index64.empty(total_length, nplike=index_nplike)

        nextcontents = []
        length_so_far = 0

        parameters = self._parameters
        for array in head:
            if isinstance(array, ak.contents.EmptyArray):
                continue

            parameters = parameters_intersect(parameters, array._parameters)
            if isinstance(array, ak.contents.UnionArray):
                union_tags = array.tags
                union_index = array.index
                union_contents = array.contents
                assert (
                    nexttags.nplike is index_nplike
                    and union_tags.nplike is index_nplike
                )
                self._backend.maybe_kernel_error(
                    self._backend[
                        "awkward_UnionArray_filltags",
                        nexttags.dtype.type,
                        union_tags.dtype.type,
                    ](
                        nexttags.data,
                        length_so_far,
                        union_tags.data,
                        array.length,
                        len(nextcontents),
                    )
                )
                assert (
                    nextindex.nplike is index_nplike
                    and union_index.nplike is index_nplike
                )
                self._backend.maybe_kernel_error(
                    self._backend[
                        "awkward_UnionArray_fillindex",
                        nextindex.dtype.type,
                        union_index.dtype.type,
                    ](
                        nextindex.data,
                        length_so_far,
                        union_index.data,
                        array.length,
                    )
                )
                length_so_far += array.length

                nextcontents.extend(union_contents)

            # This pathway is covered by `.simplified`, but we can avoid an extra array operation
            # by performing this now
            elif (
                isinstance(array, ak.contents.IndexedArray)
                and array.parameter("__array__") != "categorical"
            ):
                assert nexttags.nplike is index_nplike
                self._backend.maybe_kernel_error(
                    self._backend[
                        "awkward_UnionArray_filltags_const",
                        nexttags.dtype.type,
                    ](
                        nexttags.data,
                        length_so_far,
                        array.length,
                        len(nextcontents),
                    )
                )
                nextindex.data[
                    index_nplike.shape_item_as_index(
                        length_so_far
                    ) : index_nplike.shape_item_as_index(length_so_far + array.length)
                ] = array.index.data

                length_so_far += array.length

                nextcontents.append(
                    array.content.copy(
                        parameters=parameters_union(
                            array.content._parameters, array._parameters
                        )
                    )
                )
            else:
                assert nexttags.nplike is index_nplike
                self._backend.maybe_kernel_error(
                    self._backend[
                        "awkward_UnionArray_filltags_const",
                        nexttags.dtype.type,
                    ](
                        nexttags.data,
                        length_so_far,
                        array.length,
                        len(nextcontents),
                    )
                )

                assert nextindex.nplike is index_nplike
                self._backend.maybe_kernel_error(
                    self._backend[
                        "awkward_UnionArray_fillindex_count", nextindex.dtype.type
                    ](nextindex.data, length_so_far, array.length)
                )

                length_so_far += array.length

                nextcontents.append(array)

        if len(nextcontents) > 127:
            raise ValueError("FIXME: handle UnionArray with more than 127 contents")

        next = ak.contents.UnionArray.simplified(
            nexttags,
            nextindex,
            nextcontents,
            parameters=parameters,
        )

        # `tail` should always be empty for UnionArray's merging strategy
        assert len(tail) == 0
        return next

    def _fill_none(self, value: Content) -> Content:
        contents = []
        for content in self._contents:
            contents.append(content._fill_none(value))
        return UnionArray.simplified(
            self._tags,
            self._index,
            contents,
            parameters=self._parameters,
        )

    def _local_index(self, axis, depth):
        posaxis = maybe_posaxis(self, axis, depth)
        if posaxis is not None and posaxis + 1 == depth:
            return self._local_index_axis0()
        else:
            contents = []
            for content in self._contents:
                contents.append(content._local_index(axis, depth))
            return UnionArray(
                self._tags,
                self._index,
                contents,
                parameters=self._parameters,
            )

    def _combinations(self, n, replacement, recordlookup, parameters, axis, depth):
        posaxis = maybe_posaxis(self, axis, depth)
        if posaxis is not None and posaxis + 1 == depth:
            return self._combinations_axis0(n, replacement, recordlookup, parameters)
        else:
            contents = []
            for content in self._contents:
                contents.append(
                    content._combinations(
                        n, replacement, recordlookup, parameters, axis, depth
                    )
                )
            return ak.unionarray.UnionArray(
                self._tags,
                self._index,
                contents,
                parameters=self._parameters,
            )

    def _numbers_to_type(self, name, including_unknown):
        contents = []
        for x in self._contents:
            contents.append(x._numbers_to_type(name, including_unknown))
        return ak.contents.UnionArray(
            self._tags,
            self._index,
            contents,
            parameters=self._parameters,
        )

    def _is_unique(self, negaxis, starts, parents, outlength):
        simplified = type(self).simplified(
            self._tags,
            self._index,
            self._contents,
            parameters=self._parameters,
            mergebool=True,
        )
        if isinstance(simplified, ak.contents.UnionArray):
            raise ValueError("cannot check if an irreducible UnionArray is unique")

        return simplified._is_unique(negaxis, starts, parents, outlength)

    def _unique(self, negaxis, starts, parents, outlength):
        simplified = type(self).simplified(
            self._tags,
            self._index,
            self._contents,
            parameters=self._parameters,
            mergebool=True,
        )
        if isinstance(simplified, ak.contents.UnionArray):
            raise ValueError("cannot make a unique irreducible UnionArray")

        return simplified._unique(negaxis, starts, parents, outlength)

    def _argsort_next(
        self, negaxis, starts, shifts, parents, outlength, ascending, stable
    ):
        simplified = type(self).simplified(
            self._tags,
            self._index,
            self._contents,
            parameters=self._parameters,
            mergebool=True,
        )
        if simplified.length == 0:
            return ak.contents.NumpyArray(
                self._backend.nplike.empty(0, dtype=np.int64),
                parameters=None,
                backend=self._backend,
            )

        if isinstance(simplified, ak.contents.UnionArray):
            raise ValueError("cannot argsort an irreducible UnionArray")

        return simplified._argsort_next(
            negaxis, starts, shifts, parents, outlength, ascending, stable
        )

    def _sort_next(self, negaxis, starts, parents, outlength, ascending, stable):
        if self.length == 0:
            return self

        simplified = type(self).simplified(
            self._tags,
            self._index,
            self._contents,
            parameters=self._parameters,
            mergebool=True,
        )
        if simplified.length == 0:
            return simplified

        if isinstance(simplified, ak.contents.UnionArray):
            raise ValueError("cannot sort an irreducible UnionArray")

        return simplified._sort_next(
            negaxis, starts, parents, outlength, ascending, stable
        )

    def _reduce_next(
        self,
        reducer,
        negaxis,
        starts,
        shifts,
        parents,
        outlength,
        mask,
        keepdims,
        behavior,
    ):
        # If we have a UnionArray, it must be irreducible, thanks to the
        # canonical checks in the constructor.
        raise ValueError(f"cannot call ak.{reducer.name} on an irreducible UnionArray")

    def _validity_error(self, path):
        if self._backend.nplike.known_data and self.index.length < self.tags.length:
            return f"at {path} ({type(self)!r}): len(index) < len(tags)"

        lencontents = self._backend.index_nplike.empty(
            len(self.contents), dtype=np.int64
        )
        if self._backend.nplike.known_data:
            for j, _content_j in enumerate(self._contents):
                lencontents[j] = _content_j.length

        error = self._backend[
            "awkward_UnionArray_validity",
            self.tags.dtype.type,
            self.index.dtype.type,
            np.int64,
        ](
            self.tags.data,
            self.index.data,
            self.tags.length,
            len(self.contents),
            lencontents,
        )

        if error.str is not None:
            if error.filename is None:
                filename = ""
            else:
                filename = " (in compiled code: " + error.filename.decode(
                    errors="surrogateescape"
                ).lstrip("\n").lstrip("(")
            message = error.str.decode(errors="surrogateescape")
            return 'at {} ("{}"): {} at i={}{}'.format(
                path, type(self), message, error.id, filename
            )

        # Check triangular pairs (i, j) are not mergeable
        for i, content_i in enumerate(self._contents):
            for j in range(0, i):
                content_j = self._contents[j]
                if ak._do.mergeable(content_i, content_j, mergebool=False):
                    return f"at {path}: content({i}) is mergeable with content({j})"

        for i, content in enumerate(self._contents):
            sub = content._validity_error(path + f".content({i})")
            if sub != "":
                return sub

        return ""

    def _nbytes_part(self):
        result = self.tags._nbytes_part() + self.index._nbytes_part()
        for content in self.contents:
            result = result + content._nbytes_part()
        return result

    def _pad_none(self, target, axis, depth, clip):
        posaxis = maybe_posaxis(self, axis, depth)
        if posaxis is not None and posaxis + 1 == depth:
            return self._pad_none_axis0(target, clip)
        else:
            contents = []
            for content in self._contents:
                contents.append(content._pad_none(target, axis, depth, clip))
            return ak.contents.UnionArray.simplified(
                self.tags,
                self.index,
                contents,
                parameters=self._parameters,
            )

    def _to_arrow(self, pyarrow, mask_node, validbytes, length, options):
        nptags = self._tags.raw(numpy)
        npindex = self._index.raw(numpy)
        copied_index = False

        values = []
        for tag, content in enumerate(self._contents):
            selected_tags = nptags == tag
            this_index = npindex[selected_tags]

            # Arrow unions can't have masks; propagate validbytes down to the content.
            if validbytes is not None:
                # If this_index is a filtered permutation, we can just filter-permute
                # the mask to have the same order the content.
                if numpy.unique_values(this_index).shape[0] == this_index.shape[0]:
                    this_validbytes = numpy.zeros(this_index.shape[0], dtype=np.int8)
                    this_validbytes[this_index] = validbytes[selected_tags]

                # If this_index is not a filtered permutation, then we can't modify
                # the mask to fit the content. The same element in the content array
                # will appear multiple times in the union array, and it *might* be
                # presented as valid in some union array elements and masked in others.
                # The validbytes that recurses down to the next level can't have an
                # element that is both 0 (masked) and 1 (valid).
                else:
                    this_validbytes = validbytes[selected_tags]
                    content = content[this_index]
                    if not copied_index:
                        copied_index = True
                        npindex = numpy.asarray(npindex, copy=True)
                    npindex[selected_tags] = numpy.arange(
                        this_index.shape[0], dtype=npindex.dtype
                    )

            else:
                this_validbytes = None

            this_length = 0
            if len(this_index) != 0:
                this_length = this_index.max() + 1

            values.append(
                content._to_arrow(
                    pyarrow, mask_node, this_validbytes, this_length, options
                )
            )

        types = pyarrow.union(
            [
                pyarrow.field(str(i), values[i].type).with_nullable(
                    mask_node is not None or self._contents[i].is_option
                )
                for i in range(len(values))
            ],
            "dense",
            list(range(len(values))),
        )

        if not issubclass(npindex.dtype.type, np.int32):
            npindex = numpy.astype(npindex, dtype=np.int32)

        return pyarrow.Array.from_buffers(
            ak._connect.pyarrow.to_awkwardarrow_type(
                types,
                options["extensionarray"],
                options["record_is_scalar"],
                None,
                self,
            ),
            nptags.shape[0],
            [
                None,
                ak._connect.pyarrow.to_length(nptags, length),
                ak._connect.pyarrow.to_length(npindex, length),
            ],
            children=values,
        )

    def _to_backend_array(self, allow_missing, backend):
        raise TypeError(
            "Conversion of irreducible unions to backend arrays is not supported."
        )

    def _remove_structure(self, backend, options):
        out = []
        for i in range(len(self._contents)):
            index = self._index[self._tags.data == i]
            out.extend(
                self._contents[i]
                ._carry(index, False)
                ._remove_structure(backend, options)
            )
        return out

    def _recursively_apply(
        self, action, behavior, depth, depth_context, lateral_context, options
    ):
        if options["return_array"]:
            if options["return_simplified"]:
                make = UnionArray.simplified
            else:
                make = UnionArray

            def continuation():
                return make(
                    self._tags,
                    self._index,
                    [
                        content._recursively_apply(
                            action,
                            behavior,
                            depth,
                            copy.copy(depth_context),
                            lateral_context,
                            options,
                        )
                        for content in self._contents
                    ],
                    parameters=self._parameters if options["keep_parameters"] else None,
                )

        else:

            def continuation():
                for content in self._contents:
                    content._recursively_apply(
                        action,
                        behavior,
                        depth,
                        copy.copy(depth_context),
                        lateral_context,
                        options,
                    )

        result = action(
            self,
            depth=depth,
            depth_context=depth_context,
            lateral_context=lateral_context,
            continuation=continuation,
            behavior=behavior,
            backend=self._backend,
            options=options,
        )

        if isinstance(result, Content):
            return result
        elif result is None:
            return continuation()
        else:
            raise AssertionError(result)

    def to_packed(self) -> Self:
        tags = self._tags.raw(self._backend.nplike)
        original_index = index = self._index.raw(self._backend.nplike)[: tags.shape[0]]

        contents = list(self._contents)

        for tag in range(len(self._contents)):
            is_tag = tags == tag
            num_tag = self._backend.index_nplike.count_nonzero(is_tag)

            if len(contents[tag]) > num_tag:
                if original_index is index:
                    index = index.copy()
                index[is_tag] = self._backend.index_nplike.arange(
                    num_tag, dtype=index.dtype
                )
                contents[tag] = self.project(tag)

            contents[tag] = contents[tag].to_packed()

        return UnionArray(
            ak.index.Index8(tags, nplike=self._backend.index_nplike),
            ak.index.Index(index, nplike=self._backend.index_nplike),
            contents,
            parameters=self._parameters,
        )

    def _to_list(self, behavior, json_conversions):
        if not self._backend.nplike.known_data:
            raise TypeError("cannot convert typetracer arrays to Python lists")

        out = self._to_list_custom(behavior, json_conversions)
        if out is not None:
            return out

        tags = self._tags.raw(numpy)
        index = self._index.raw(numpy)
        contents = [x._to_list(behavior, json_conversions) for x in self._contents]

        out = [None] * tags.shape[0]
        for i, tag in enumerate(tags):
            out[i] = contents[tag][index[i]]
        return out

    def _to_backend(self, backend: Backend) -> Self:
        tags = self._tags.to_nplike(backend.index_nplike)
        index = self._index.to_nplike(backend.index_nplike)
        contents = [content.to_backend(backend) for content in self._contents]
        return UnionArray(
            tags,
            index,
            contents,
            parameters=self._parameters,
        )

    def _is_equal_to(self, other, index_dtype, numpyarray):
        return (
            self.tags.is_equal_to(other.tags)
            and self.index.is_equal_to(other.index, index_dtype, numpyarray)
            and len(self.contents) == len(other.contents)
            and all(
                self.contents[i].is_equal_to(other.contents[i], index_dtype, numpyarray)
                for i in range(len(self.contents))
            )
        )
